# Copyright 2013 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from builtins import bytes
import logging
import struct

from chirp import chirp_common, directory, memmap, errors, util
from chirp import bitwise
from chirp.settings import RadioSettingGroup, RadioSetting
from chirp.settings import RadioSettingValueBoolean, RadioSettingValueList
from chirp.settings import RadioSettingValueString, RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0030;
struct {
  lbcd rx_freq[4];
  lbcd tx_freq[4];
  ul16 rx_tone;
  ul16 tx_tone;
  u8 signaling:2,
     unknown1:3,
     bcl:1,
     wide:1,
     beatshift:1;
  u8 pttid:2,
     highpower:1,
     scan:1,
     unknown2:4;
  u8 unknown3[2];
} memory[8];

#seekto 0x0250;
struct {
  u8 unknown0;
  u8 unknown1;
  u8 unknown2;
  u8 unknown3;
  // 0x04 Emergency, 0x05 keylock, 0x06 Monitor 0x09 temp delete, 0x0A none
  u8 speakerkey;
  u8 circlekey;
} settings;

#seekto 0x0270;
struct {
  u8 scanresume; // 0=selected 1=selected+talkback
  u8 dropoutdelay; // 0.5s*value, 0.5-5.0
  u8 txdwell; // same as above
  u8 offhookscan; // 0=off 1=on
} settings2;

#seekto 0x0310;
struct {
  char line1[32];
  char line2[32];
} messages;

"""

# These are 4+index
KEYS = [
    'Emergency',
    'Key Lock',
    'Monitor',
    'Scan On/Off',
    'Talk Around',
    'Temporary Delete',
    'None',
]
SCAN_RESUME = ['Selected', 'Selected+Talkback']
SCAN_TIME = ['%.1f' % (v * 0.5) for v in range(11)]
POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=5),
                chirp_common.PowerLevel("High", watts=50)]
MODES = ["NFM", "FM"]
PTTID = ["", "BOT", "EOT", "Both"]
SIGNAL = ["", "DTMF"]


def make_frame(cmd, addr, length, data=b""):
    return struct.pack(">BHB", ord(cmd), addr, length) + bytes(data)


def send(radio, frame):
    # LOG.debug("%04i P>R: %s" % (len(frame), util.hexprint(frame)))
    radio.pipe.write(frame)


def recv(radio, readdata=True):
    hdr = radio.pipe.read(4)
    cmd, addr, length = struct.unpack(">BHB", hdr)
    if readdata:
        data = radio.pipe.read(length)
        # LOG.debug("     P<R: %s" % util.hexprint(hdr + data))
        if len(data) != length:
            raise errors.RadioError("Radio sent %i bytes (expected %i)" % (
                    len(data), length))
    else:
        data = bytes(b"")
    radio.pipe.write(b"\x06")
    return addr, data


def do_ident(radio):
    send(radio, b"PROGRAM")
    ack = radio.pipe.read(1)
    if ack != bytes(b"\x06"):
        LOG.debug('Response was %r, expected 0x06' % ack)
        raise errors.RadioError("Radio refused program mode")
    radio.pipe.write(b"\x02")
    ident = radio.pipe.read(8)
    try:
        modelstr = ident[1:5].decode()
    except UnicodeDecodeError:
        LOG.debug('Model string was %r' % ident)
        modelstr = '?'
    if modelstr != radio.MODEL.split("-")[1]:
        raise errors.RadioError("Incorrect model: TK-%s, expected %s" % (
                modelstr, radio.MODEL))
    LOG.info("Model: %s" % util.hexprint(ident))
    radio.pipe.write(b"\x06")
    ack = radio.pipe.read(1)


def do_download(radio):
    radio.pipe.parity = "E"
    radio.pipe.timeout = 1
    try:
        do_ident(radio)
    except errors.RadioError:
        do_ident(radio)

    data = bytes(b"")
    for addr in range(0, 0x0400, 8):
        send(radio, make_frame(bytes(b"R"), addr, 8))
        _addr, _data = recv(radio)
        if _addr != addr:
            raise errors.RadioError("Radio sent unexpected address")
        data += _data
        radio.pipe.write(b"\x06")
        ack = radio.pipe.read(1)
        if ack != bytes(b"\x06"):
            raise errors.RadioError("Radio refused block at %04x" % addr)

        status = chirp_common.Status()
        status.cur = addr
        status.max = 0x0400
        status.msg = "Cloning from radio"
        radio.status_fn(status)

    radio.pipe.write(b"\x45")

    data = (b"\x45\x58\x33\x34\x30\x32\xff\xff" + (b"\xff" * 8) +
            data)
    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    radio.pipe.parity = "E"
    radio.pipe.timeout = 1
    try:
        do_ident(radio)
    except errors.RadioError:
        do_ident(radio)

    mmap = radio._mmap.get_byte_compatible()
    for addr in range(0, 0x0400, 8):
        eaddr = addr + 16
        send(radio, make_frame(b"W", addr, 8, mmap[eaddr:eaddr + 8]))
        ack = radio.pipe.read(1)
        if ack != bytes(b"\x06"):
            raise errors.RadioError("Radio refused block at %04x" % addr)
        radio.pipe.write(b"\x06")

        status = chirp_common.Status()
        status.cur = addr
        status.max = 0x0400
        status.msg = "Cloning to radio"
        radio.status_fn(status)

    radio.pipe.write(b"\x45")


class KenwoodTKx102Radio(chirp_common.CloneModeRadio):
    """Kenwood TK-x102"""
    VENDOR = "Kenwood"
    MODEL = "TK-x102"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False

    _memsize = 0x410

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_cross = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.has_name = False
        rf.has_rx_dtcs = True
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_modes = MODES
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_skips = ["", "S"]
        rf.valid_bands = [self._range]
        rf.memory_bounds = (1, self._upper)
        return rf

    def sync_in(self):
        try:
            self._mmap = do_download(self)
        except errors.RadioError:
            self.pipe.write(b"\x45")
            raise
        except Exception as e:
            raise errors.RadioError("Failed to download from radio: %s" % e)
        self.process_mmap()

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_out(self):
        try:
            do_upload(self)
        except errors.RadioError:
            self.pipe.write(b"\x45")
            raise
        except Exception as e:
            raise errors.RadioError("Failed to upload to radio: %s" % e)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def _get_tone(self, _mem, mem):
        def _get_dcs(val):
            code = int("%03o" % (val & 0x07FF))
            pol = (val & 0x8000) and "R" or "N"
            return code, pol

        if _mem.tx_tone != 0xFFFF and _mem.tx_tone > 0x2800:
            tcode, tpol = _get_dcs(_mem.tx_tone)
            mem.dtcs = tcode
            txmode = "DTCS"
        elif _mem.tx_tone != 0xFFFF:
            mem.rtone = _mem.tx_tone / 10.0
            txmode = "Tone"
        else:
            txmode = ""

        if _mem.rx_tone != 0xFFFF and _mem.rx_tone > 0x2800:
            rcode, rpol = _get_dcs(_mem.rx_tone)
            mem.rx_dtcs = rcode
            rxmode = "DTCS"
        elif _mem.rx_tone != 0xFFFF:
            mem.ctone = _mem.rx_tone / 10.0
            rxmode = "Tone"
        else:
            rxmode = ""

        if txmode == "Tone" and not rxmode:
            mem.tmode = "Tone"
        elif txmode == rxmode and txmode == "Tone" and mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif txmode == rxmode and txmode == "DTCS" and mem.dtcs == mem.rx_dtcs:
            mem.tmode = "DTCS"
        elif rxmode or txmode:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txmode, rxmode)

        if mem.tmode == "DTCS":
            mem.dtcs_polarity = "%s%s" % (tpol, rpol)

    def get_memory(self, number):
        _mem = self._memobj.memory[number - 1]

        mem = chirp_common.Memory()
        mem.number = number

        if _mem.get_raw(asbytes=False)[:4] == "\xFF\xFF\xFF\xFF":
            mem.empty = True
            return mem

        mem.freq = int(_mem.rx_freq) * 10
        offset = (int(_mem.tx_freq) * 10) - mem.freq
        if offset < 0:
            mem.offset = abs(offset)
            mem.duplex = "-"
        elif offset > 0:
            mem.offset = offset
            mem.duplex = "+"
        else:
            mem.offset = 0

        self._get_tone(_mem, mem)
        mem.power = POWER_LEVELS[_mem.highpower]
        mem.mode = MODES[_mem.wide]
        mem.skip = not _mem.scan and "S" or ""

        mem.extra = RadioSettingGroup("all", "All Settings")

        bcl = RadioSetting("bcl", "Busy Channel Lockout",
                           RadioSettingValueBoolean(bool(_mem.bcl)))
        mem.extra.append(bcl)

        beat = RadioSetting("beatshift", "Beat Shift",
                            RadioSettingValueBoolean(bool(_mem.beatshift)))
        mem.extra.append(beat)

        pttid = RadioSetting("pttid", "PTT ID",
                             RadioSettingValueList(PTTID,
                                                   PTTID[_mem.pttid]))
        mem.extra.append(pttid)

        signal = RadioSetting("signaling", "Signaling",
                              RadioSettingValueList(SIGNAL,
                                                    SIGNAL[
                                                      _mem.signaling & 0x01]))
        mem.extra.append(signal)

        return mem

    def _set_tone(self, mem, _mem):
        def _set_dcs(code, pol):
            val = int("%i" % code, 8) + 0x2800
            if pol == "R":
                val += 0xA000
            return val

        rx_mode = tx_mode = None
        rx_tone = tx_tone = 0xFFFF

        if mem.tmode == "Tone":
            tx_mode = "Tone"
            rx_mode = None
            tx_tone = int(mem.rtone * 10)
        elif mem.tmode == "TSQL":
            rx_mode = tx_mode = "Tone"
            rx_tone = tx_tone = int(mem.ctone * 10)
        elif mem.tmode == "DTCS":
            tx_mode = rx_mode = "DTCS"
            tx_tone = _set_dcs(mem.dtcs, mem.dtcs_polarity[0])
            rx_tone = _set_dcs(mem.dtcs, mem.dtcs_polarity[1])
        elif mem.tmode == "Cross":
            tx_mode, rx_mode = mem.cross_mode.split("->")
            if tx_mode == "DTCS":
                tx_tone = _set_dcs(mem.dtcs, mem.dtcs_polarity[0])
            elif tx_mode == "Tone":
                tx_tone = int(mem.rtone * 10)
            if rx_mode == "DTCS":
                rx_tone = _set_dcs(mem.rx_dtcs, mem.dtcs_polarity[1])
            elif rx_mode == "Tone":
                rx_tone = int(mem.ctone * 10)

        _mem.rx_tone = rx_tone
        _mem.tx_tone = tx_tone

        LOG.debug("Set TX %s (%i) RX %s (%i)" %
                  (tx_mode, _mem.tx_tone, rx_mode, _mem.rx_tone))

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]

        if mem.empty:
            _mem.set_raw("\xFF" * 16)
            return

        _mem.unknown3[0] = 0x07
        _mem.unknown3[1] = 0x22
        _mem.rx_freq = mem.freq / 10
        if mem.duplex == "+":
            _mem.tx_freq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.tx_freq = (mem.freq - mem.offset) / 10
        else:
            _mem.tx_freq = mem.freq / 10

        self._set_tone(mem, _mem)

        _mem.highpower = mem.power == POWER_LEVELS[1]
        _mem.wide = mem.mode == "FM"
        _mem.scan = mem.skip != "S"

        for setting in mem.extra:
            if setting.get_name == "signaling":
                if setting.value == "DTMF":
                    _mem.signaling = 0x03
                else:
                    _mem.signaling = 0x00
            else:
                setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        _mem = self._memobj
        basic = RadioSettingGroup("basic", "Basic")
        scan = RadioSettingGroup("scan", "Scan")
        top = RadioSettings(basic, scan)

        def _f(val):
            string = ""
            for char in str(val):
                if char == "\xFF":
                    break
                string += char
            return string

        line1 = RadioSetting("messages.line1", "Message Line 1",
                             RadioSettingValueString(0, 32,
                                                     _f(_mem.messages.line1),
                                                     autopad=False))
        basic.append(line1)

        line2 = RadioSetting("messages.line2", "Message Line 2",
                             RadioSettingValueString(0, 32,
                                                     _f(_mem.messages.line2),
                                                     autopad=False))
        basic.append(line2)

        skey = RadioSetting(
            "settings.speakerkey", "Speaker Key",
            RadioSettingValueList(
                KEYS, current_index=int(_mem.settings.speakerkey) - 4))
        basic.append(skey)

        ckey = RadioSetting(
            "settings.circlekey", "Circle Key",
            RadioSettingValueList(
                KEYS, current_index=int(_mem.settings.circlekey) - 4))
        basic.append(ckey)

        scanresume = RadioSetting(
            "settings2.scanresume", "Scan Resume",
            RadioSettingValueList(SCAN_RESUME,
                                  current_index=_mem.settings2.scanresume))
        scan.append(scanresume)

        offhookscan = RadioSetting(
            "settings2.offhookscan", "Off-Hook Scan",
            RadioSettingValueBoolean(bool(_mem.settings2.offhookscan)))
        scan.append(offhookscan)

        dropout = RadioSetting(
            "settings2.dropoutdelay", "Drop-out Delay Time",
            RadioSettingValueList(
                SCAN_TIME, current_index=int(_mem.settings2.dropoutdelay)))

        scan.append(dropout)

        txdwell = RadioSetting(
            "settings2.txdwell", "TX Dwell Time",
            RadioSettingValueList(
                SCAN_TIME, current_index=int(_mem.settings2.txdwell)))
        scan.append(txdwell)

        return top

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if "." in element.get_name():
                bits = element.get_name().split(".")
                obj = self._memobj
                for bit in bits[:-1]:
                    obj = getattr(obj, bit)
                setting = bits[-1]
            else:
                obj = self._memobj.settings
                setting = element.get_name()

            if "line" in setting:
                value = str(element.value).ljust(32, "\xFF")
            elif 'key' in setting:
                value = int(element.value) + 4
            else:
                value = element.value
            setattr(obj, setting, value)

    @classmethod
    def match_model(cls, filedata, filename):
        model = filedata[0x03D1:0x03D5]
        return model == cls.MODEL.encode().split(b"-")[1]


@directory.register
class KenwoodTK7102Radio(KenwoodTKx102Radio):
    MODEL = "TK-7102"
    _range = (136000000, 174000000)
    _upper = 4


@directory.register
class KenwoodTK8102Radio(KenwoodTKx102Radio):
    MODEL = "TK-8102"
    _range = (400000000, 500000000)
    _upper = 4


@directory.register
class KenwoodTK7108Radio(KenwoodTKx102Radio):
    MODEL = "TK-7108"
    _range = (136000000, 174000000)
    _upper = 8


@directory.register
class KenwoodTK8108Radio(KenwoodTKx102Radio):
    MODEL = "TK-8108"
    _range = (400000000, 500000000)
    _upper = 8



# Copyright 2022 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import queue
import threading
import uuid

LOG = logging.getLogger(__name__)
_JOB_COUNTER = 0
_JOB_COUNTER_LOCK = threading.Lock()


def jobnumber():
    global _JOB_COUNTER

    # This shouldn't be called outside the main thread, so this should
    # not be necessary. But, be careful anyway in case that changes in
    # the future.
    with _JOB_COUNTER_LOCK:
        num = _JOB_COUNTER
        _JOB_COUNTER += 1
    return num


class RadioJob:
    def __init__(self, editor, fn, args, kwargs):
        self.editor = editor
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.result = None
        self.id = str(uuid.uuid4())
        self.jobnumber = jobnumber()

    @property
    def score(self):
        if self.fn == 'get_memory':
            return 20
        elif self.fn.startswith('get'):
            return 10
        else:
            return 0

    def __lt__(self, job):
        # Prioritize jobs with lower scores and jobs submitted before us
        return self.score < job.score or self.jobnumber < job.jobnumber

    def __repr__(self):
        return '<%s@%i>%s(%s,%s)=%r' % (
            self.__class__.__name__,
            self.score,
            self.fn,
            ','.join(repr(a) for a in self.args),
            ','.join('%s=%r' % (k, v) for k, v in self.kwargs.items()),
            self.result)

    def dispatch(self, radio):
        try:
            self.result = getattr(radio, self.fn)(*self.args, **self.kwargs)
        except Exception as e:
            LOG.exception('Failed to run %r' % self)
            self.result = e

        LOG.debug('Radio finished %r' % self)


class BackgroundRadioJob(RadioJob):
    @property
    def score(self):
        return 100


class RadioThread(threading.Thread):
    SENTINEL = RadioJob(None, 'END', [], {})

    def __init__(self, radio):
        super().__init__()
        self._radio = radio
        self._queue = queue.PriorityQueue()
        self._log = logging.getLogger('RadioThread')
        self._waiting = []

    def submit(self, editor, fn, *a, **k):
        job = RadioJob(editor, fn, a, k)
        self._queue.put(job)
        return job.id

    def background(self, editor, fn, *a, **k):
        job = BackgroundRadioJob(editor, fn, a, k)
        self._queue.put(job)
        return job.id

    def end(self):
        self._queue.put(self.SENTINEL)

    def run(self):
        while True:
            job = self._queue.get()
            if job is self.SENTINEL:
                self._log.info('Exiting on request')
                return
            job.dispatch(self._radio)
            self._waiting.append(job)

            for job in list(self._waiting):
                delivered = job.editor.radio_thread_event(
                    job, block=not self.pending)
                if delivered:
                    self._waiting.remove(job)

    @property
    def pending(self):
        return self._queue.qsize()



# Copyright 2013 Sean Burford <sburford@google.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import bandplan

SHORTNAME = "iaru_r2"

DESC = {
  "name": "IARU Region 2 (The Americas)",
  "updated": "October 8, 2010",
  "url": "http://www.iaru.org/uploads/1/3/0/7/13073366/r2_band_plan.pdf"
}

# Bands are broken up like this so that other plans can import bits.

BANDS_160M = (
  bandplan.Band((1800000, 2000000), "160 Meter Band"),
  bandplan.Band((1800000, 1810000), "Digimodes"),
  bandplan.Band((1810000, 1830000), "CW", mode="CW"),
  bandplan.Band((1830000, 1840000), "CW, priority for DX", mode="CW"),
  bandplan.Band((1840000, 1850000), "SSB, priority for DX", mode="LSB"),
  bandplan.Band((1850000, 1999000), "All modes", mode="LSB"),
  bandplan.Band((1999000, 2000000), "Beacons", mode="CW"),
)

BANDS_80M = (
  bandplan.Band((3500000, 4000000), "80 Meter Band"),
  bandplan.Band((3500000, 3510000), "CW, priority for DX", mode="CW"),
  bandplan.Band((3510000, 3560000), "CW, contest preferred", mode="CW"),
  bandplan.Band((3560000, 3580000), "CW", mode="CW"),
  bandplan.Band((3580000, 3590000), "All narrow band modes, digimodes"),
  bandplan.Band((3590000, 3600000), "All modes"),
  bandplan.Band((3600000, 3650000), "All modes, SSB contest preferred",
                mode="LSB"),
  bandplan.Band((3650000, 3700000), "All modes", mode="LSB"),
  bandplan.Band((3700000, 3775000), "All modes, SSB contest preferred",
                mode="LSB"),
  bandplan.Band((3775000, 3800000), "All modes, SSB DX preferred", mode="LSB"),
  bandplan.Band((3800000, 4000000), "All modes"),
)

BANDS_40M = (
  bandplan.Band((7000000, 7300000), "40 Meter Band"),
  bandplan.Band((7000000, 7025000), "CW, priority for DX", mode="CW"),
  bandplan.Band((7025000, 7035000), "CW", mode="CW"),
  bandplan.Band((7035000, 7038000), "All narrow band modes, digimodes"),
  bandplan.Band((7038000, 7040000), "All narrow band modes, digimodes"),
  bandplan.Band((7040000, 7043000), "All modes, digimodes"),
  bandplan.Band((7043000, 7300000), "All modes"),
)

BANDS_30M = (
  bandplan.Band((10100000, 10150000), "30 Meter Band"),
  bandplan.Band((10100000, 10130000), "CW", mode="CW"),
  bandplan.Band((10130000, 10140000), "All narrow band digimodes"),
  bandplan.Band((10140000, 10150000), "All modes, digimodes, no phone"),
)

BANDS_20M = (
  bandplan.Band((14000000, 14350000), "20 Meter Band"),
  bandplan.Band((14000000, 14025000), "CW, priority for DX", mode="CW"),
  bandplan.Band((14025000, 14060000), "CW, contest preferred", mode="CW"),
  bandplan.Band((14060000, 14070000), "CW", mode="CW"),
  bandplan.Band((14070000, 14089000), "All narrow band modes, digimodes"),
  bandplan.Band((14089000, 14099000), "All modes, digimodes"),
  bandplan.Band((14099000, 14101000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((14101000, 14112000), "All modes, digimodes"),
  bandplan.Band((14112000, 14285000), "All modes, SSB contest preferred",
                mode="USB"),
  bandplan.Band((14285000, 14300000), "All modes", mode="AM"),
  bandplan.Band((14300000, 14350000), "All modes"),
)

BANDS_17M = (
  bandplan.Band((18068000, 18168000), "17 Meter Band"),
  bandplan.Band((18068000, 18095000), "CW", mode="CW"),
  bandplan.Band((18095000, 18105000), "All narrow band modes, digimodes"),
  bandplan.Band((18105000, 18109000), "All narrow band modes, digimodes"),
  bandplan.Band((18109000, 18111000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((18111000, 18120000), "All modes, digimodes"),
  bandplan.Band((18120000, 18168000), "All modes"),
)

BANDS_15M = (
  bandplan.Band((21000000, 21450000), "15 Meter Band"),
  bandplan.Band((21000000, 21070000), "CW", mode="CW"),
  bandplan.Band((21070000, 21090000), "All narrow band modes, digimodes"),
  bandplan.Band((21090000, 21110000), "All narrow band modes, digimodes"),
  bandplan.Band((21110000, 21120000), "All modes (exc SSB), digimodes"),
  bandplan.Band((21120000, 21149000), "All narrow band modes"),
  bandplan.Band((21149000, 21151000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((21151000, 21450000), "All modes", mode="USB"),
)

BANDS_12M = (
  bandplan.Band((24890000, 24990000), "12 Meter Band"),
  bandplan.Band((24890000, 24915000), "CW", mode="CW"),
  bandplan.Band((24915000, 24925000), "All narrow band modes, digimodes"),
  bandplan.Band((24925000, 24929000), "All narrow band modes, digimodes"),
  bandplan.Band((24929000, 24931000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((24931000, 24940000), "All modes, digimodes"),
  bandplan.Band((24940000, 24990000), "All modes", mode="USB"),
)

BANDS_10M = (
  bandplan.Band((28000000, 29520000), "10 Meter Band"),
  bandplan.Band((28000000, 28070000), "CW", mode="CW"),
  bandplan.Band((28070000, 28120000), "All narrow band modes, digimodes"),
  bandplan.Band((28120000, 28150000), "All narrow band modes, digimodes"),
  bandplan.Band((28150000, 28190000), "All narrow band modes, digimodes"),
  bandplan.Band((28190000, 28199000), "Regional time shared beacons",
                mode="CW"),
  bandplan.Band((28199000, 28201000), "IBP, exclusively for beacons",
                mode="CW"),
  bandplan.Band((28201000, 28225000), "Continuous duty beacons",
                mode="CW"),
  bandplan.Band((28225000, 28300000), "All modes, beacons"),
  bandplan.Band((28300000, 28320000), "All modes, digimodes"),
  bandplan.Band((28320000, 29000000), "All modes"),
  bandplan.Band((29000000, 29200000), "All modes, AM preferred", mode="AM"),
  bandplan.Band((29200000, 29300000), "All modes including FM, digimodes"),
  bandplan.Band((29300000, 29510000), "Satellite downlink"),
  bandplan.Band((29510000, 29520000), "Guard band, no transmission allowed"),
  bandplan.Band((29520000, 29700000), "FM", step_khz=10, mode="NFM"),
  bandplan.Band((29620000, 29690000), "FM Repeaters", input_offset=-100000),
)

BANDS = BANDS_160M + BANDS_80M + BANDS_40M + BANDS_30M + BANDS_20M + \
        BANDS_17M + BANDS_15M + BANDS_12M + BANDS_10M



# Copyright 2021 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0010;
struct {
    lbcd rxfreq[4];
    lbcd txfreq[4];
    lbcd rxtone[2];
    lbcd txtone[2];
    u8 speccode:1,       // Retevis RB87 PTT ID
       compander:1,
       scramble:1,
       skip:1,
       highpower:1,
       narrow:1,
       unknown2:1,
       bcl:1;
    u8 unknown3[3];
} memory[%d];
#seekto 0x02B0;
struct {
    u8 voicesw;      // Voice SW            +
    u8 voiceselect;  // Voice Select
    u8 scan;         // Scan                +
    u8 vox;          // VOX                 +
    u8 voxgain;      // Vox Gain            +
    u8 voxnotxonrx;  // Rx Disable Vox      +
    u8 hivoltnotx;   // High Vol Inhibit TX +
    u8 lovoltnotx;   // Low Vol Inhibit TX  +
    u8 rxemergency;  // RX Emergency
} settings2;
#seekto 0x03C0;
struct {
    u8 codesw:1,         // Retevis RB29 code switch
       scanmode:1,
       vox:1,            // Retevis RB19 VOX
       speccode:1,
       voiceprompt:2,
       batterysaver:1,
       beep:1;           // Retevis RB87 vox
    u8 squelchlevel;
    u8 sidekey2;         // Retevis RT22S setting
                         // Retevis RB85 sidekey 1 short
                         // Retevis RB19 sidekey 2 long
                         // Retevis RT47 sidekey 1 long
                         // Retevis RB87 sidekey 1 long
    u8 timeouttimer;
    u8 voxlevel;
    u8 sidekey2S;
    u8 unused;           // Selected channel
    u8 voxdelay;
    u8 sidekey1L;
    u8 sidekey2L;
    u8 unused2[3];
    u8 unknown3:4,
       unknown4:1,
       unknown5:2,
       power10w:1;       // Retevis RT85 power 10w on/off
                         // Retevis RT75 stop TX with low voltage
} settings;
"""

MEM_FORMAT_RB18 = """
#seekto 0x0000;
struct {
    lbcd rxfreq[4];
    lbcd txfreq[4];
    lbcd rxtone[2];
    lbcd txtone[2];
    u8 jumpcode:1,
       unknown1:2,
       skip:1,
       highpower:1,
       narrow:1,
       unknown2:1,
       bcl:1;
    u8 unknown3[3];
} memory[%d];
#seekto 0x0630;
struct {
    u8 unk630:7,
       voice:1;
    u8 unk631:7,
       language:1;
    u8 unk632:7,
       scan:1;
    u8 unk633:7,
       vox:1;
    u8 unk634:5,
       vox_level:3;
    u8 unk635;
    u8 unk636:7,
       lovoltnotx:1;
    u8 unk637:7,
       hivoltnotx:1;
    u8 unknown2[8];
    u8 unk640:5,
       rogerbeep:1,
       batterysaver:1,
       beep:1;
    u8 squelchlevel;
    u8 unk642;
    u8 timeouttimer;
    u8 unk644:7,
       tail:1;
    u8 channel;
} settings;
"""

MEM_FORMAT_T20FRS = """
// #seekto 0x0000;
struct {
    lbcd rxfreq[4];
    lbcd txfreq[4];
    lbcd rxtone[2];
    lbcd txtone[2];
    u8 jumpcode:1,
       unknown1:2,
       skip:1,
       highpower:1,
       narrow:1,
       unknown2:1,
       bcl:1;
    u8 unknown3[3];
} memory[%d];
#seekto 0x02B0;
struct {
    u8 voicesw;      // Voice SW            +
    u8 voiceselect;  // Voice Select
    u8 scan;         // Scan                +
    u8 vox;          // VOX                 +
    u8 voxgain;      // Vox Gain            +
    u8 voxnotxonrx;  // Rx Disable Vox      +
    u8 hivoltnotx;   // High Vol Inhibit TX +
    u8 lovoltnotx;   // Low Vol Inhibit TX  +
    u8 rxemergency;  // RX Emergency
} settings2;
#seekto 0x02C0;
struct {
    u8 unk:6,
       batterysaver:1,
       beep:1;
    u8 squelchlevel;
    u8 sidekey2;
    u8 timeouttimer;
} settings;
"""

CMD_ACK = b"\x06"

VOICE_LIST = ["Off", "Chinese", "English"]
VOICE_LIST2 = ["English", "Chinese"]
VOICE_LIST3 = ["Off", "English", "Chinese"]
VOICE_LIST4 = ["Chinese", "English"]
TIMEOUTTIMER_LIST = ["Off", "30 seconds", "60 seconds", "90 seconds",
                     "120 seconds", "150 seconds", "180 seconds",
                     "210 seconds", "240 seconds", "270 seconds",
                     "300 seconds"]
SCANMODE_LIST = ["Carrier", "Time"]
VOXLEVEL_LIST = ["Off", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
VOXDELAY_LIST = ["0.5 seconds", "1.0 seconds", "1.5 seconds",
                 "2.0 seconds", "2.5 seconds", "3.0 seconds"]
SIDEKEY19_LIST = ["Off", "Scan", "Emergency Alarm"]
SIDEKEY2_LIST = SIDEKEY19_LIST + ["Display Battery"]
SIDEKEY29_LIST = ["Off", "Scan", "VOX", "Busy Lock", "Emergency Alarm"]

SIDEKEY85SHORT_LIST = ["Off",
                       "Noise Cansellation On",
                       "Continuous Monitor",
                       "High/Low Power",
                       "Emergency Alarm",
                       "Show Battery",
                       "Scan",
                       "VOX",
                       "Busy Channel Lock"]
SIDEKEY85LONG_LIST = ["Off",
                      "Noise Cansellation On",
                      "Continuous Monitor",
                      "Monitor Momentary",
                      "High/Low Power",
                      "Emergency Alarm",
                      "Show Battery",
                      "Scan",
                      "VOX",
                      "Busy Channel Lock"]
SPECCODE_LIST = ["SpeCode 1", "SpeCode 2"]
SIDEKEY75_LIST = ["Off",
                  "Monitor Momentary",
                  "Scan",
                  "VOX",
                  "Monitor",
                  "Announciation"]
SIDEKEY47_LIST = ["Monitor Momentary",
                  "Channel Lock",
                  "Scan",
                  "VOX"]
SIDEKEYV8A_LIST = ["Off",
                   "Monitor",
                   "High/Low Power",
                   "Alarm"]
SIDEKEY87_LIST = ["Scan", "Emergency Alarm"]

FRS_FREQS1 = [462562500, 462587500, 462612500, 462637500, 462662500,
              462687500, 462712500]
FRS_FREQS2 = [467562500, 467587500, 467612500, 467637500, 467662500,
              467687500, 467712500]
FRS_FREQS3 = [462550000, 462575000, 462600000, 462625000, 462650000,
              462675000, 462700000, 462725000]
FRS_FREQS = FRS_FREQS1 + FRS_FREQS2 + FRS_FREQS3

FRS16_FREQS = [462562500, 462587500, 462612500, 462637500,
               462662500, 462625000, 462725000, 462687500,
               462712500, 462550000, 462575000, 462600000,
               462650000, 462675000, 462700000, 462725000]

GMRS_FREQS = FRS_FREQS1 + FRS_FREQS2 + FRS_FREQS3 * 2

MURS_FREQS = [151820000, 151880000, 151940000, 154570000, 154600000]

PMR_FREQS1 = [446006250, 446018750, 446031250, 446043750, 446056250,
              446068750, 446081250, 446093750]
PMR_FREQS2 = [446106250, 446118750, 446131250, 446143750, 446156250,
              446168750, 446181250, 446193750]
PMR_FREQS = PMR_FREQS1 + PMR_FREQS2


def _t18_enter_programming_mode(radio):
    serial = radio.pipe

    _magic = b"\x02" + radio._magic

    try:
        serial.write(_magic)
        if radio._echo:
            chew = serial.read(len(_magic))  # Chew the echo
        ack = serial.read(1)
    except:
        raise errors.RadioError("Error communicating with radio")

    if not ack:
        raise errors.RadioError("No response from radio")
    elif ack != CMD_ACK:
        raise errors.RadioError("Radio refused to enter programming mode")

    try:
        serial.write(b"\x02")
        if radio._echo:
            serial.read(1)  # Chew the echo
        ident = serial.read(8)
    except:
        raise errors.RadioError("Error communicating with radio")

    # check if ident is OK
    for fp in radio._fingerprint:
        if ident.startswith(fp):
            break
    else:
        LOG.debug("Incorrect model ID, got this:\n\n" + util.hexprint(ident))
        raise errors.RadioError("Radio identification failed.")

    try:
        serial.write(CMD_ACK)
        if radio._echo:
            serial.read(1)  # Chew the echo
        ack = serial.read(1)
    except:
        raise errors.RadioError("Error communicating with radio")

    if radio.MODEL == "RT647":
        if ack != b"\xF0":
            raise errors.RadioError("Radio refused to enter programming mode")
    else:
        if ack != CMD_ACK:
            raise errors.RadioError("Radio refused to enter programming mode")


def _t18_exit_programming_mode(radio):
    serial = radio.pipe
    try:
        serial.write(radio.CMD_EXIT)
        if radio._echo:
            chew = serial.read(1)  # Chew the echo
    except:
        raise errors.RadioError("Radio refused to exit programming mode")


def _t18_read_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'R', block_addr, block_size)
    expectedresponse = b"W" + cmd[1:]
    LOG.debug("Reading block %04x..." % (block_addr))

    try:
        serial.write(cmd)
        if radio._echo:
            serial.read(4)  # Chew the echo
        response = serial.read(4 + block_size)
        if response[:4] != expectedresponse:
            raise Exception("Error reading block %04x." % (block_addr))

        block_data = response[4:]

        if radio.ACK_BLOCK:
            serial.write(CMD_ACK)
            if radio._echo:
                serial.read(1)  # Chew the echo
            ack = serial.read(1)
    except:
        raise errors.RadioError("Failed to read block at %04x" % block_addr)

    if radio.ACK_BLOCK:
        if ack != CMD_ACK:
            raise Exception("No ACK reading block %04x." % (block_addr))

    return block_data


def _t18_write_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'W', block_addr, block_size)
    data = radio.get_mmap()[block_addr:block_addr + block_size]

    LOG.debug("Writing Data:")
    LOG.debug(util.hexprint(cmd + data))

    try:
        serial.write(cmd + data)
        if radio._echo:
            serial.read(4 + len(data))  # Chew the echo
        if serial.read(1) != CMD_ACK:
            raise Exception("No ACK")
    except:
        raise errors.RadioError("Failed to send block "
                                "to radio at %04x" % block_addr)


def do_download(radio):
    LOG.debug("download")
    _t18_enter_programming_mode(radio)

    data = b""

    status = chirp_common.Status()
    status.msg = "Cloning from radio"

    status.cur = 0
    status.max = radio._memsize

    for addr in range(0, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr + radio.BLOCK_SIZE
        radio.status_fn(status)

        block = _t18_read_block(radio, addr, radio.BLOCK_SIZE)
        data += block

        LOG.debug("Address: %04x" % addr)
        LOG.debug(util.hexprint(block))

    _t18_exit_programming_mode(radio)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"

    _t18_enter_programming_mode(radio)

    status.cur = 0
    status.max = radio._memsize

    for start_addr, end_addr in radio._ranges:
        for addr in range(start_addr, end_addr, radio.BLOCK_SIZE):
            status.cur = addr + radio.BLOCK_SIZE
            radio.status_fn(status)
            _t18_write_block(radio, addr, radio.BLOCK_SIZE)

    _t18_exit_programming_mode(radio)


def model_match(cls, data):
    """Match the opened/downloaded image to the correct version"""

    if len(data) == cls._memsize:
        rid = data[0x03D0:0x03D8]
        return b"P558" in rid
    else:
        return False


@directory.register
class T18Radio(chirp_common.CloneModeRadio):
    """radtel T18"""
    VENDOR = "Radtel"
    MODEL = "T18"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False
    BLOCK_SIZE = 0x08
    CMD_EXIT = b"b"
    ACK_BLOCK = True

    VALID_BANDS = [(400000000, 470000000)]

    _magic = b"1ROGRAM"
    _fingerprint = [b"SMP558" + b"\x00\x00"]
    _upper = 16
    _mem_params = (_upper  # number of channels
                   )
    _frs = _frs16 = _murs = _pmr = _gmrs = False
    _echo = False

    _ranges = [
        (0x0000, 0x03F0),
    ]
    _memsize = 0x03F0

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.valid_modes = ["NFM", "FM"]  # 12.5 kHz, 25 kHz.
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        if self.MODEL != "T18" and \
                self.MODEL != "RB618" and \
                self.MODEL != "RT647":
            rf.valid_power_levels = self.POWER_LEVELS
        rf.can_odd_split = True
        rf.has_rx_dtcs = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.has_tuning_step = False
        rf.has_bank = False
        rf.has_name = False
        rf.memory_bounds = (1, self._upper)
        rf.valid_bands = self.VALID_BANDS
        rf.valid_tuning_steps = chirp_common.TUNING_STEPS

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT % self._mem_params, self._mmap)

    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    def sync_out(self):
        do_upload(self)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def _decode_tone(self, val):
        val = int(val)
        if val == 16665:
            return '', None, None
        elif val >= 12000:
            return 'DTCS', val - 12000, 'R'
        elif val >= 8000:
            return 'DTCS', val - 8000, 'N'
        else:
            return 'Tone', val / 10.0, None

    def _encode_tone(self, memval, mode, value, pol):
        if mode == '':
            memval[0].set_raw(0xFF)
            memval[1].set_raw(0xFF)
        elif mode == 'Tone':
            memval.set_value(int(value * 10))
        elif mode == 'DTCS':
            flag = 0x80 if pol == 'N' else 0xC0
            memval.set_value(value)
            memval[1].set_bits(flag)
        else:
            raise Exception("Internal error: invalid mode `%s'" % mode)

    def get_memory(self, number):
        _mem = self._memobj.memory[number - 1]

        mem = chirp_common.Memory()

        mem.number = number
        mem.freq = int(_mem.rxfreq) * 10

        # We'll consider any blank (i.e. 0 MHz frequency) to be empty
        if mem.freq == 0:
            mem.empty = True
            return mem

        if _mem.rxfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            mem.freq = 0
            mem.empty = True
            return mem

        if _mem.txfreq.get_raw() == b"\xFF\xFF\xFF\xFF":
            mem.duplex = "off"
            mem.offset = 0
        elif int(_mem.rxfreq) == int(_mem.txfreq):
            mem.duplex = ""
            mem.offset = 0
        else:
            mem.duplex = int(_mem.rxfreq) > int(_mem.txfreq) and "-" or "+"
            mem.offset = abs(int(_mem.rxfreq) - int(_mem.txfreq)) * 10

        mem.mode = not _mem.narrow and "FM" or "NFM"

        mem.skip = _mem.skip and "S" or ""

        txtone = self._decode_tone(_mem.txtone)
        rxtone = self._decode_tone(_mem.rxtone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        if self.MODEL != "T18" and self.MODEL != "RB618":
            mem.power = self.POWER_LEVELS[1 - _mem.highpower]

        mem.extra = RadioSettingGroup("Extra", "extra")
        rs = RadioSetting("bcl", "Busy Channel Lockout",
                          RadioSettingValueBoolean(not _mem.bcl))
        mem.extra.append(rs)
        if self.MODEL != "RB18" and self.MODEL != "RB618" and \
                self.MODEL != "FRS-B1" and self.MODEL != "BF-V8A" and \
                self.MODEL != "RB29" and self.MODEL != "RB629" and \
                self.MODEL != "BF-T20FRS":
            if self.MODEL not in ["RB87",
                                  "RT47V"]:
                rs = RadioSetting("scramble", "Scramble",
                                  RadioSettingValueBoolean(not _mem.scramble))
                mem.extra.append(rs)
            rs = RadioSetting("compander", "Compander",
                              RadioSettingValueBoolean(not _mem.compander))
            mem.extra.append(rs)
            if self.MODEL != "RT47V" and self.MODEL != "T8" and \
                    self.MODEL != "RB17" and self.MODEL != "RB617" and \
                    self.MODEL != "RB75" and self.MODEL != "RB19P":
                if self.MODEL == "RB87":
                    rs = RadioSettingValueBoolean(not _mem.speccode)
                    rset = RadioSetting("speccode", "PTT-ID", rs)
                    mem.extra.append(rset)
                else:
                    rs = RadioSettingValueBoolean(not _mem.speccode)
                    rset = RadioSetting("speccode", "Spec Code", rs)
                    mem.extra.append(rset)

        immutable = []

        if self._frs:
            if mem.freq in FRS_FREQS:
                if mem.number >= 1 and mem.number <= 22:
                    FRS_FREQ = FRS_FREQS[mem.number - 1]
                    mem.freq = FRS_FREQ
                mem.duplex = ''
                mem.offset = 0
                mem.mode = "NFM"
                if mem.number >= 8 and mem.number <= 14:
                    mem.power = self.POWER_LEVELS[1]
                    immutable = ["empty", "freq", "duplex", "offset", "mode",
                                 "power"]
                else:
                    immutable = ["empty", "freq", "duplex", "offset", "mode"]
        elif self._frs16:
            if mem.freq in FRS16_FREQS:
                if mem.number >= 1 and mem.number <= 16:
                    FRS_FREQ = FRS16_FREQS[mem.number - 1]
                    mem.freq = FRS_FREQ
                mem.duplex = ''
                mem.offset = 0
                mem.mode = "NFM"
                immutable = ["empty", "freq", "duplex", "offset", "mode"]
        elif self._murs:
            if mem.freq in MURS_FREQS:
                if mem.number >= 1 and mem.number <= 5:
                    MURS_FREQ = MURS_FREQS[mem.number - 1]
                    mem.freq = MURS_FREQ
                mem.duplex = ''
                mem.offset = 0
                if mem.number <= 3:
                    mem.mode = "NFM"
                    immutable = ["empty", "freq", "duplex", "offset", "mode"]
                else:
                    immutable = ["empty", "freq", "duplex", "offset"]
        elif self._pmr:
            if mem.freq in PMR_FREQS:
                if mem.number >= 1 and mem.number <= 16:
                    PMR_FREQ = PMR_FREQS[mem.number - 1]
                    mem.freq = PMR_FREQ
                mem.duplex = ''
                mem.offset = 0
                mem.mode = "NFM"
                mem.power = self.POWER_LEVELS[1]
                immutable = ["empty", "freq", "duplex", "offset", "mode",
                             "power"]
        elif self._gmrs:
            if mem.freq in GMRS_FREQS:
                if mem.number >= 1 and mem.number <= 30:
                    GMRS_FREQ = GMRS_FREQS[mem.number - 1]
                    mem.freq = GMRS_FREQ
                    immutable = ["freq"]
                if mem.number >= 1 and mem.number <= 7:
                    mem.duplex = ''
                    mem.offset = 0
                    immutable += ["duplex", "offset"]
                elif mem.number >= 8 and mem.number <= 14:
                    mem.duplex = ''
                    mem.offset = 0
                    mem.mode = "NFM"
                    mem.power = self.POWER_LEVELS[1]
                    immutable += ["duplex", "offset", "mode", "power"]
                elif mem.number >= 15 and mem.number <= 22:
                    mem.duplex = ''
                    mem.offset = 0
                    immutable += ["duplex", "offset"]
                elif mem.number >= 23 and mem.number <= 30:
                    mem.duplex = '+'
                    mem.offset = 5000000
                    immutable += ["duplex", "offset"]

        mem.immutable = immutable

        return mem

    def set_memory(self, mem):
        # Get a low-level memory object mapped to the image
        _mem = self._memobj.memory[mem.number - 1]

        if mem.empty:
            _mem.set_raw("\xFF" * (_mem.size() // 8))

            return

        if self.MODEL == "BF-V8A" or self.MODEL == "BF-T20FRS":
            _mem.set_raw("\x00" * 12 + "\x09\xFF\xFF\xFF")
        elif self.MODEL == "RB29" or self.MODEL == "RB629":
            _mem.set_raw("\x00" * 12 + "\xF7\xFF\xFF\xFF")
        else:
            _mem.set_raw("\x00" * 12 + "\xF9\xFF\xFF\xFF")

        _mem.rxfreq = mem.freq / 10

        if mem.duplex == "off":
            for i in range(0, 4):
                _mem.txfreq[i].set_raw("\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

        txtone, rxtone = chirp_common.split_tone_encode(mem)
        self._encode_tone(_mem.txtone, *txtone)
        self._encode_tone(_mem.rxtone, *rxtone)

        if self.MODEL != "T18" and self.MODEL != "RB18":
            _mem.highpower = mem.power == self.POWER_LEVELS[0]

        _mem.narrow = 'N' in mem.mode
        _mem.skip = mem.skip == "S"

        for setting in mem.extra:
            # NOTE: Only three settings right now, all are inverted
            setattr(_mem, setting.get_name(), not int(setting.value))

    def get_settings(self):
        _settings = self._memobj.settings
        if self.MODEL == "FRS-B1" or self.MODEL == "BF-V8A" or \
                self.MODEL == "BF-T20FRS":
            _settings2 = self._memobj.settings2
        basic = RadioSettingGroup("basic", "Basic Settings")
        top = RadioSettings(basic)

        rs = RadioSetting("squelchlevel", "Squelch level",
                          RadioSettingValueInteger(
                              0, 9, _settings.squelchlevel))
        basic.append(rs)

        rs = RadioSetting("timeouttimer", "Timeout timer",
                          RadioSettingValueList(
                              TIMEOUTTIMER_LIST,
                              TIMEOUTTIMER_LIST[
                                  _settings.timeouttimer]))
        basic.append(rs)

        if self.MODEL == "RB18" or self.MODEL == "RB618":
            rs = RadioSetting("scan", "Scan",
                              RadioSettingValueBoolean(_settings.scan))
            basic.append(rs)
        elif self.MODEL == "FRS-B1" or self.MODEL == "BF-V8A" or \
                self.MODEL == "BF-T20FRS":
            rs = RadioSetting("settings2.scan", "Scan",
                              RadioSettingValueBoolean(_settings2.scan))
            basic.append(rs)
        else:
            rs = RadioSetting("scanmode", "Scan mode",
                              RadioSettingValueList(
                                  SCANMODE_LIST,
                                  SCANMODE_LIST[_settings.scanmode]))
            basic.append(rs)

        if self.MODEL == "RT20":
            rs = RadioSetting("voiceprompt", "Voice prompts",
                              RadioSettingValueList(
                                  VOICE_LIST4,
                                  VOICE_LIST4[_settings.voiceprompt]))
            basic.append(rs)
        elif self.MODEL == "RT22S":
            rs = RadioSetting("voiceprompt", "Voice prompts",
                              RadioSettingValueBoolean(_settings.voiceprompt))
            basic.append(rs)
        elif self.MODEL == "RB18" or self.MODEL == "RB618":
            rs = RadioSetting("voice", "Voice prompts",
                              RadioSettingValueBoolean(_settings.voice))
            basic.append(rs)
        elif self.MODEL == "FRS-B1" or self.MODEL == "BF-V8A" or \
                self.MODEL == "BF-T20FRS":
            rs = RadioSetting("settings2.voicesw", "Voice prompts",
                              RadioSettingValueBoolean(_settings2.voicesw))
            basic.append(rs)
        elif self.MODEL == "RT15":
            rs = RadioSetting("voiceprompt", "Voice prompts",
                              RadioSettingValueList(
                                  VOICE_LIST3,
                                  VOICE_LIST3[_settings.voiceprompt]))
            basic.append(rs)
        else:
            if self.MODEL != "RB87":
                rs = RadioSetting("voiceprompt", "Voice prompts",
                                  RadioSettingValueList(
                                      VOICE_LIST,
                                      VOICE_LIST[_settings.voiceprompt]))
                basic.append(rs)

        rs = RadioSetting("batterysaver", "Battery saver",
                          RadioSettingValueBoolean(_settings.batterysaver))
        basic.append(rs)

        if self.MODEL not in ["RB29",
                              "RB75",
                              "RB87",
                              "RB629",
                              "RT15"
                              ]:
            rs = RadioSetting("beep", "Beep",
                              RadioSettingValueBoolean(_settings.beep))
            basic.append(rs)

        if self.MODEL == "RB19" or self.MODEL == "RB19P" \
                or self.MODEL == "RB619":
            rs = RadioSetting("vox", "VOX",
                              RadioSettingValueBoolean(_settings.vox))
            basic.append(rs)

        if self.MODEL == "RB87":
            rs = RadioSetting("beep", "VOX",
                              RadioSettingValueBoolean(_settings.beep))
            basic.append(rs)

        if self.MODEL != "RB18" and self.MODEL != "RB618" \
                and self.MODEL != "FRS-B1" \
                and self.MODEL != "BF-V8A" \
                and self.MODEL != "BF-T20FRS":
            rs = RadioSetting("voxlevel", "Vox level",
                              RadioSettingValueList(
                                  VOXLEVEL_LIST,
                                  VOXLEVEL_LIST[_settings.voxlevel]))
            basic.append(rs)

            rs = RadioSetting("voxdelay", "VOX delay",
                              RadioSettingValueList(
                                  VOXDELAY_LIST,
                                  VOXDELAY_LIST[_settings.voxdelay]))
            basic.append(rs)

        if self.MODEL == "RT22S":
            rs = RadioSetting("sidekey2", "Side Key 2(Long)",
                              RadioSettingValueList(
                                  SIDEKEY2_LIST,
                                  SIDEKEY2_LIST[_settings.sidekey2]))
            basic.append(rs)

        if self.MODEL == "RB18" or self.MODEL == "RB618":
            rs = RadioSetting("language", "Language",
                              RadioSettingValueList(
                                  VOICE_LIST2,
                                  VOICE_LIST2[_settings.language]))
            basic.append(rs)

            rs = RadioSetting("tail", "Tail",
                              RadioSettingValueBoolean(_settings.tail))
            basic.append(rs)

            rs = RadioSetting("hivoltnotx", "High voltage no TX",
                              RadioSettingValueBoolean(_settings.hivoltnotx))
            basic.append(rs)

            rs = RadioSetting("lovoltnotx", "Low voltage no TX",
                              RadioSettingValueBoolean(_settings.lovoltnotx))
            basic.append(rs)

            rs = RadioSetting("vox", "VOX",
                              RadioSettingValueBoolean(_settings.vox))
            basic.append(rs)

            if _settings.vox_level > 4:
                val = 1
            else:
                val = _settings.vox_level + 1
            rs = RadioSetting("vox_level", "VOX level",
                              RadioSettingValueInteger(1, 5, val))
            basic.append(rs)

            rs = RadioSetting("rogerbeep", "Roger beep",
                              RadioSettingValueBoolean(_settings.rogerbeep))
            basic.append(rs)

        if self.MODEL == "RB85":
            rs = RadioSetting("speccode", "SpecCode Select",
                              RadioSettingValueList(
                                  SPECCODE_LIST,
                                  SPECCODE_LIST[_settings.speccode]))
            basic.append(rs)

            rs = RadioSetting("sidekey2", "Side Key 1(Short)",
                              RadioSettingValueList(
                                  SIDEKEY85SHORT_LIST,
                                  SIDEKEY85SHORT_LIST[_settings.sidekey2]))
            basic.append(rs)

            rs = RadioSetting("sidekey1L", "Side Key 1(Long)",
                              RadioSettingValueList(
                                  SIDEKEY85LONG_LIST,
                                  SIDEKEY85LONG_LIST[_settings.sidekey1L]))
            basic.append(rs)

            rs = RadioSetting("sidekey2S", "Side Key 2(Short)",
                              RadioSettingValueList(
                                  SIDEKEY85SHORT_LIST,
                                  SIDEKEY85SHORT_LIST[_settings.sidekey2S]))
            basic.append(rs)

            rs = RadioSetting("sidekey2L", "Side Key 2(Long)",
                              RadioSettingValueList(
                                  SIDEKEY85LONG_LIST,
                                  SIDEKEY85LONG_LIST[_settings.sidekey2L]))
            basic.append(rs)

            rs = RadioSetting("power10w", "Power 10W",
                              RadioSettingValueBoolean(_settings.power10w))
            basic.append(rs)

        if self.MODEL == "RB75":
            rs = RadioSetting("sidekey2", "Side Key 1(Short)",
                              RadioSettingValueList(
                                  SIDEKEY75_LIST,
                                  SIDEKEY75_LIST[_settings.sidekey2]))
            basic.append(rs)

            rs = RadioSetting("sidekey1L", "Side Key 1(Long)",
                              RadioSettingValueList(
                                  SIDEKEY75_LIST,
                                  SIDEKEY75_LIST[_settings.sidekey1L]))
            basic.append(rs)

            rs = RadioSetting("sidekey2S", "Side Key 2(Short)",
                              RadioSettingValueList(
                                  SIDEKEY75_LIST,
                                  SIDEKEY75_LIST[_settings.sidekey2S]))
            basic.append(rs)

            rs = RadioSetting("sidekey2L", "Side Key 2(Long)",
                              RadioSettingValueList(
                                  SIDEKEY75_LIST,
                                  SIDEKEY75_LIST[_settings.sidekey2L]))
            basic.append(rs)

            rs = RadioSetting("power10w", "Low Voltage Stop TX",
                              RadioSettingValueBoolean(_settings.power10w))
            basic.append(rs)

        if self.MODEL == "RB87":
            rs = RadioSetting("sidekey2", "Side Key 1(Long)",
                              RadioSettingValueList(
                                  SIDEKEY87_LIST,
                                  SIDEKEY87_LIST[_settings.sidekey2]))
            basic.append(rs)

        if self.MODEL == "FRS-B1":
            rs = RadioSetting("settings2.hivoltnotx",
                              "High Voltage Inhibit TX",
                              RadioSettingValueBoolean(_settings2.hivoltnotx))
            basic.append(rs)

            rs = RadioSetting("settings2.lovoltnotx", "Low Voltage Inhibit TX",
                              RadioSettingValueBoolean(_settings2.lovoltnotx))
            basic.append(rs)

            rs = RadioSetting("settings2.vox", "Vox",
                              RadioSettingValueBoolean(_settings2.vox))
            basic.append(rs)

            rs = RadioSetting("settings2.voxnotxonrx", "Rx Disable VOX",
                              RadioSettingValueBoolean(_settings2.voxnotxonrx))
            basic.append(rs)

            rs = RadioSetting("settings2.voxgain", "Vox Gain",
                              RadioSettingValueInteger(
                                  1, 5, _settings2.voxgain))
            basic.append(rs)

        if self.MODEL == "RB19" or self.MODEL == "RB19P" \
                or self.MODEL == "RB619":
            rs = RadioSetting("sidekey2", "Left Navigation Button(Long)",
                              RadioSettingValueList(
                                  SIDEKEY19_LIST,
                                  SIDEKEY19_LIST[_settings.sidekey2]))
            basic.append(rs)

        if self.MODEL == "RT47" or self.MODEL == "RT47V" or \
                self.MODEL == "RT647":
            rs = RadioSetting("sidekey2", "Side Key 1(Long)",
                              RadioSettingValueList(
                                  SIDEKEY47_LIST,
                                  SIDEKEY47_LIST[_settings.sidekey2]))
            basic.append(rs)

            rs = RadioSetting("sidekey2S", "Side Key 2(Long)",
                              RadioSettingValueList(
                                  SIDEKEY47_LIST,
                                  SIDEKEY47_LIST[_settings.sidekey2S]))
            basic.append(rs)

        if self.MODEL == "BF-V8A" or self.MODEL == "BF-T20FRS":
            rs = RadioSetting("sidekey2", "Side key",
                              RadioSettingValueList(
                                  SIDEKEYV8A_LIST,
                                  SIDEKEYV8A_LIST[_settings.sidekey2]))
            basic.append(rs)

            rs = RadioSetting("settings2.rxemergency", "RX emergency",
                              RadioSettingValueBoolean(_settings2.rxemergency))
            basic.append(rs)

            rs = RadioSetting("settings2.voiceselect", "Language",
                              RadioSettingValueList(
                                  VOICE_LIST2,
                                  VOICE_LIST2[_settings2.voiceselect]))
            basic.append(rs)

            rs = RadioSetting("settings2.hivoltnotx",
                              "High voltage inhibit TX",
                              RadioSettingValueBoolean(_settings2.hivoltnotx))
            basic.append(rs)

            rs = RadioSetting("settings2.lovoltnotx", "Low voltage inhibit TX",
                              RadioSettingValueBoolean(_settings2.lovoltnotx))
            basic.append(rs)

            rs = RadioSetting("settings2.vox", "VOX",
                              RadioSettingValueBoolean(_settings2.vox))
            basic.append(rs)

            rs = RadioSetting("settings2.voxnotxonrx", "RX disable VOX",
                              RadioSettingValueBoolean(_settings2.voxnotxonrx))
            basic.append(rs)

            rs = RadioSetting("settings2.voxgain", "Vox Gain",
                              RadioSettingValueInteger(
                                  1, 5, _settings2.voxgain))
            basic.append(rs)

        if self.MODEL == "RB29" or self.MODEL == "RB629":
            rs = RadioSetting("codesw", "Code Switch",
                              RadioSettingValueBoolean(_settings.codesw))
            basic.append(rs)

            rs = RadioSetting("sidekey2S", "Side Key(Short)",
                              RadioSettingValueList(
                                  SIDEKEY29_LIST,
                                  SIDEKEY29_LIST[_settings.sidekey2S]))
            basic.append(rs)

            rs = RadioSetting("sidekey2L", "Side Key(Long)",
                              RadioSettingValueList(
                                  SIDEKEY29_LIST,
                                  SIDEKEY29_LIST[_settings.sidekey2L]))
            basic.append(rs)

        return top

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        bits = element.get_name().split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif setting == "vox_level":
                        setattr(obj, setting, int(element.value) - 1)
                    else:
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        if cls.MODEL == "T18":
            match_size = False
            match_model = False

            # testing the file data size
            if len(filedata) == cls._memsize:
                match_size = True

            # testing the model fingerprint
            match_model = model_match(cls, filedata)

            if match_size and match_model:
                return True
            else:
                return False
        else:
            # Radios that have always been post-metadata, so never do
            # old-school detection
            return False


@directory.register
class RT20Radio(T18Radio):
    """RETEVIS RT20"""
    VENDOR = "Retevis"
    MODEL = "RT20"
    ACK_BLOCK = True
    BLOCK_SIZE = 0x08

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low",  watts=0.50)]

    _magic = b"8AOGRAM"
    _fingerprint = [b"SMP558" + b"\x02"]
    _upper = 16
    _mem_params = (_upper  # number of channels
                   )


@directory.register
class RT22SRadio(T18Radio):
    """RETEVIS RT22S"""
    VENDOR = "Retevis"
    MODEL = "RT22S"
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low",  watts=0.50)]

    _magic = b"9COGRAM"
    _fingerprint = [b"SMP558" + b"\x02"]
    _upper = 22
    _mem_params = (_upper  # number of channels
                   )
    _frs = True
    _pmr = False


@directory.register
class RB18Radio(T18Radio):
    """RETEVIS RB18"""
    VENDOR = "Retevis"
    MODEL = "RB18"
    BLOCK_SIZE = 0x10
    CMD_EXIT = b"E"

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low",  watts=0.50)]

    _magic = b"PROGRAL"
    _fingerprint = [b"P3107" + b"\xF7"]
    _upper = 22
    _mem_params = (_upper  # number of channels
                   )
    _frs = True
    _pmr = False

    _ranges = [
        (0x0000, 0x0660),
    ]
    _memsize = 0x0660

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT_RB18 %
                                     self._mem_params, self._mmap)

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class RB618Radio(RB18Radio):
    """RETEVIS RB618"""
    VENDOR = "Retevis"
    MODEL = "RB618"

    _upper = 16
    _mem_params = (_upper  # number of channels
                   )
    _frs = False
    _pmr = True


@directory.register
class RT68Radio(T18Radio):
    """RETEVIS RT68"""
    VENDOR = "Retevis"
    MODEL = "RT68"
    ACK_BLOCK = False
    CMD_EXIT = b""

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low",  watts=0.50)]

    _magic = b"83OGRAM"
    _fingerprint = [b"\x06\x00\x00\x00\x00\x00\x00\x00"]
    _upper = 16
    _mem_params = (_upper  # number of channels
                   )
    _frs16 = True
    _pmr = False

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class RT668Radio(RT68Radio):
    """RETEVIS RT668"""
    VENDOR = "Retevis"
    MODEL = "RT668"

    _frs16 = False
    _pmr = True


@directory.register
class RB17Radio(RT68Radio):
    """RETEVIS RB17"""
    VENDOR = "Retevis"
    MODEL = "RB17"

    _magic = b"A5OGRAM"
    _fingerprint = [b"\x53\x00\x00\x00\x00\x00\x00\x00"]

    _frs16 = True
    _pmr = False
    _murs = False


@directory.register
class RB617Radio(RB17Radio):
    """RETEVIS RB617"""
    VENDOR = "Retevis"
    MODEL = "RB617"

    _frs16 = False
    _pmr = True
    _murs = False


@directory.register
class RB17VRadio(RB17Radio):
    """RETEVIS RB17V"""
    VENDOR = "Retevis"
    MODEL = "RB17V"

    VALID_BANDS = [(136000000, 174000000)]

    _upper = 5

    _frs16 = False
    _pmr = False
    _murs = True


@directory.register
class RB85Radio(T18Radio):
    """Retevis RB85"""
    VENDOR = "Retevis"
    MODEL = "RB85"
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=10.00),
                    chirp_common.PowerLevel("Low", watts=5.00)]

    _magic = b"H19GRAM"
    _fingerprint = [b"SMP558" + b"\x02"]


@directory.register
class RB75Radio(T18Radio):
    """Retevis RB75"""
    VENDOR = "Retevis"
    MODEL = "RB75"
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _magic = b"KVOGRAM"
    _fingerprint = [b"SMP558" + b"\x00"]
    _upper = 30
    _mem_params = (_upper  # number of channels
                   )
    _gmrs = False  # sold as GMRS radio but supports full band TX/RX


@directory.register
class FRSB1Radio(T18Radio):
    """BTECH FRS-B1"""
    VENDOR = "BTECH"
    MODEL = "FRS-B1"
    ACK_BLOCK = True

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _magic = b"PROGRAM"
    _fingerprint = [b"P3107" + b"\xF7\x00"]
    _upper = 22
    _mem_params = (_upper  # number of channels
                   )
    _frs = True


@directory.register
class RB19Radio(T18Radio):
    """Retevis RB19"""
    VENDOR = "Retevis"
    MODEL = "RB19"
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _magic = b"9COGRAM"
    _fingerprint = [b"SMP558" + b"\x02"]
    _upper = 22
    _mem_params = (_upper  # number of channels
                   )
    _frs = True


@directory.register
class RB19PRadio(T18Radio):
    """Retevis RB19P"""
    VENDOR = "Retevis"
    MODEL = "RB19P"
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=3.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _magic = b"70OGRAM"
    _fingerprint = [b"SMP558" + b"\x02"]
    _upper = 30
    _mem_params = (_upper  # number of channels
                   )
    _gmrs = True


@directory.register
class RB619Radio(T18Radio):
    """Retevis RB619"""
    VENDOR = "Retevis"
    MODEL = "RB619"
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=0.500),
                    chirp_common.PowerLevel("Low", watts=0.499)]

    _magic = b"9COGRAM"
    _fingerprint = [b"SMP558" + b"\x02"]
    _upper = 16
    _mem_params = (_upper  # number of channels
                   )
    _pmr = True


@directory.register
class RT47Radio(T18Radio):
    """Retevis RT47"""
    VENDOR = "Retevis"
    MODEL = "RT47"
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.000),
                    chirp_common.PowerLevel("Low", watts=0.500)]

    _magic = b"47OGRAM"
    _fingerprint = [b"\x06\x00\x00\x00\x00\x00\x00\x00"]
    _upper = 16
    _mem_params = (_upper  # number of channels
                   )
    _frs16 = True
    _echo = True


@directory.register
class RT47VRadio(RT47Radio):
    """Retevis RT47V"""
    VENDOR = "Retevis"
    MODEL = "RT47V"

    VALID_BANDS = [(136000000, 174000000)]

    _upper = 5
    _mem_params = (_upper  # number of channels
                   )
    _frs16 = False
    _murs = True


@directory.register
class RT647Radio(RT47Radio):
    """Retevis RT647"""
    VENDOR = "Retevis"
    MODEL = "RT647"

    _frs16 = False
    _pmr = True


@directory.register
class BFV8ARadio(T18Radio):
    """Baofeng BF-V8A"""
    VENDOR = "Baofeng"
    MODEL = "BF-V8A"
    ACK_BLOCK = True

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.000),
                    chirp_common.PowerLevel("Low", watts=0.500)]

    _magic = b"PROGRAM"
    _fingerprint = [b"P3107" + b"\xF7\x00\x00"]
    _upper = 16
    _mem_params = (_upper  # number of channels
                   )
    _echo = False


@directory.register
class RB29Radio(T18Radio):
    """Retevis RB29"""
    VENDOR = "Retevis"
    MODEL = "RB29"
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _magic = b"S19GRAM"
    _fingerprint = [b"SMP558" + b"\x02"]
    _upper = 16
    _mem_params = (_upper  # number of channels
                   )
    _frs16 = True


@directory.register
class RB629Radio(RB29Radio):
    """Retevis RB29"""
    VENDOR = "Retevis"
    MODEL = "RB629"

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=0.500),
                    chirp_common.PowerLevel("Low", watts=0.499)]

    _frs16 = False
    _pmr = True


@directory.register
class RT15Radio(T18Radio):
    """RETEVIS RT15"""
    VENDOR = "Retevis"
    MODEL = "RT15"
    ACK_BLOCK = False
    CMD_EXIT = b"b"

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.00),
                    chirp_common.PowerLevel("Low",  watts=0.50)]

    _magic = b"KAOGRAM"
    _fingerprint = [b"\x06\x00\x00\x00\x00\x00\x00\x00",
                    b"\x06\x03\xE8\x08\xFF\xFF\xFF\xFF"]
    _upper = 16
    _mem_params = (_upper  # number of channels
                   )
    _frs16 = False  # sold as FRS radio but supports full band TX/RX

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
class RB87Radio(T18Radio):
    """RETEVIS RB87"""
    VENDOR = "Retevis"
    MODEL = "RB87"
    ACK_BLOCK = False
    CMD_EXIT = b""
    ACK_BLOCK = False

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5.00),
                    chirp_common.PowerLevel("Low", watts=0.50)]

    _magic = b"C8OGRAN"
    _fingerprint = [b"SMP558"]
    _upper = 30
    _mem_params = (_upper  # number of channels
                   )
    _gmrs = True


@directory.register
class BFT20FRSRadio(T18Radio):
    """Baofeng BF-T20FRS"""
    VENDOR = "Baofeng"
    MODEL = "BF-T20FRS"
    ACK_BLOCK = True

    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=2.000),
                    chirp_common.PowerLevel("Low", watts=0.500)]

    _magic = b"PROGRAM"
    _fingerprint = [b"P3107" + b"\xF7\x00\x00"]
    _upper = 22
    _mem_params = (_upper  # number of channels
                   )
    _frs = False  # sold as FRS radio but supports full band TX/RX

    _ranges = [
        (0x0000, 0x0160),
        (0x02B0, 0x02D0),
    ]
    _memsize = 0x03F0

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT_T20FRS %
                                     self._mem_params, self._mmap)



# Copyright 2012 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import icf
from chirp import directory, bitwise, chirp_common, settings

LOG = logging.getLogger(__name__)
MEM_FORMAT = """
struct {
  u24 freq;
  u16 offset;
  u16 rtone:6,
      ctone:6,
      unknown2:1,
      mode:3;
  u8 dtcs;
  u8 tune_step:4,
     dsql_mode:2,
     unknown5:2;
  u8 unknown4;
  u8 tmode:4,
     duplex:2,
     dtcs_polarity:2;
  char name[16];
  u8 digcode;
  u8 urcall[7];
  u8 rpt1call[7];
  u8 rpt2call[7];
} memory[500];

#seekto 0x69C0;
u8 used_flags[70];

#seekto 0x6A06;
u8 skip_flags[69];

#seekto 0x6A4B;
u8 pskp_flags[69];

#seekto 0x6AC0;
struct {
  u8 bank;
  u8 index;
} banks[500];

#seekto 0x6F50;
struct {
  char name[16];
} bank_names[26];

#seekto 0x74BF;
struct {
  u8 unknown0;
  u24 freq;
  u16 offset;
  u8 unknown1[3];
  u8 call[7];
  char name[16];
  char subname[8];
  u8 unknown3[9];
} repeaters[700];

#seekto 0xFABC;
struct {
  u8 call[7];
} rptcall[700];

#seekto 0x10F20;
struct {
  char call[8];
  char tag[4];
} mycall[6];

#seekto 0x10F68;
struct {
  char call[8];
} urcall[200];

"""

TMODES = ["", "Tone", "TSQL", "TSQL", "DTCS", "DTCS", "TSQL-R", "DTCS-R"]
DUPLEX = ["", "-", "+"]
DTCS_POLARITY = ["NN", "NR", "RN", "RR"]
TUNING_STEPS = [5.0, 6.25, 0, 0, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0,
                100.0, 125.0, 200.0]
DSQL_MODES = ['', 'DSQL', 'CSQL']


def _decode_call(_call):
    # Why Icom, why?
    return ''.join(chr(x) for x in icf.warp_byte_size(_call, 7, 8))


def _encode_call(call):
    return list(icf.warp_byte_size(call.ljust(8), 8, 7))


def _get_freq(_mem):
    freq = int(_mem.freq)
    offs = int(_mem.offset)

    if freq & 0x00200000:
        mult = 6250
    else:
        mult = 5000

    freq &= 0x0003FFFF

    return (freq * mult), (offs * mult)


def _set_freq(_mem, freq, offset):
    if chirp_common.is_fractional_step(freq):
        mult = 6250
        flag = 0x00200000
    else:
        mult = 5000
        flag = 0x00000000

    _mem.freq = (freq // mult) | flag
    _mem.offset = (offset // mult)


class ID31Bank(icf.IcomBank):
    """A ID-31 Bank"""
    def get_name(self):
        _banks = self._model._radio._memobj.bank_names
        return str(_banks[self.index].name).rstrip()

    def set_name(self, name):
        _banks = self._model._radio._memobj.bank_names
        _banks[self.index].name = str(name).ljust(16)[:16]


@directory.register
class ID31Radio(icf.IcomCloneModeRadio, chirp_common.IcomDstarSupport):
    """Icom ID-31"""
    MODEL = "ID-31A"

    _memsize = 0x15500
    _model = "\x33\x22\x00\x01"
    _endframe = "Icom Inc\x2E\x41\x38"
    _num_banks = 26
    _bank_class = ID31Bank
    _can_hispeed = True

    _icf_data = {
        'MapRev': 1,
        'EtcData': 0x000005,
    }

    _ranges = [(0x00000, 0x15500, 32)]

    MODES = {0: "FM", 1: "NFM", 5: "DV"}

    def _get_bank(self, loc):
        _bank = self._memobj.banks[loc]
        if _bank.bank == 0xFF:
            return None
        else:
            return _bank.bank

    def _set_bank(self, loc, bank):
        _bank = self._memobj.banks[loc]
        if bank is None:
            _bank.bank = 0xFF
        else:
            _bank.bank = bank

    def _get_bank_index(self, loc):
        _bank = self._memobj.banks[loc]
        return _bank.index

    def _set_bank_index(self, loc, index):
        _bank = self._memobj.banks[loc]
        _bank.index = index

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (0, 499)
        rf.valid_bands = [(400000000, 479000000)]
        rf.has_settings = True
        rf.has_ctone = True
        rf.has_bank_index = True
        rf.has_bank_names = True
        rf.valid_tmodes = list(TMODES)
        rf.valid_tuning_steps = sorted(list(TUNING_STEPS))
        rf.valid_modes = list(self.MODES.values())
        rf.valid_skips = ["", "S", "P"]
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.valid_name_length = 16
        rf.requires_call_lists = False
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def _get_mem_extra(self, _mem, mem):
        extra = settings.RadioSettingGroup('extra', 'Extra')
        rs = settings.RadioSetting(
            'dsql', 'DSQL Mode',
            settings.RadioSettingValueList(
                DSQL_MODES, current_index=int(_mem.dsql_mode)))
        extra.append(rs)
        return extra

    def _set_mem_extra(self, _mem, mem):
        try:
            _mem.dsql_mode = DSQL_MODES.index(str(mem.extra['dsql'].value))
        except (KeyError, TypeError):
            # No settings or dsql is not provided
            pass
        except ValueError:
            LOG.error('Invalid extra.dsql_mode: %r' % str(
                mem.extra['dsql'].value))

    def get_memory(self, number):
        _mem = self._memobj.memory[number]
        _usd = self._memobj.used_flags[number / 8]
        _skp = self._memobj.skip_flags[number / 8]
        _psk = self._memobj.pskp_flags[number / 8]

        bit = (1 << (number % 8))

        if self.MODES[int(_mem.mode)] == "DV":
            mem = chirp_common.DVMemory()
        else:
            mem = chirp_common.Memory()
        mem.number = number

        if _usd & bit:
            mem.empty = True
            return mem

        mem.freq, mem.offset = _get_freq(_mem)
        mem.name = str(_mem.name).rstrip()
        mem.rtone = chirp_common.TONES[_mem.rtone]
        mem.ctone = chirp_common.TONES[_mem.ctone]
        mem.tmode = TMODES[_mem.tmode]
        mem.duplex = DUPLEX[_mem.duplex]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.dtcs_polarity = DTCS_POLARITY[_mem.dtcs_polarity]
        mem.tuning_step = TUNING_STEPS[_mem.tune_step]
        mem.mode = self.MODES[int(_mem.mode)]

        if mem.mode == "DV":
            mem.dv_urcall = _decode_call(_mem.urcall).rstrip()
            mem.dv_rpt1call = _decode_call(_mem.rpt1call).rstrip()
            mem.dv_rpt2call = _decode_call(_mem.rpt2call).rstrip()
            mem.dv_code = int(_mem.digcode)

        if _psk & bit:
            mem.skip = "P"
        elif _skp & bit:
            mem.skip = "S"

        mem.extra = self._get_mem_extra(_mem, mem)

        return mem

    def set_memory(self, memory):
        _mem = self._memobj.memory[memory.number]
        _usd = self._memobj.used_flags[memory.number / 8]
        _skp = self._memobj.skip_flags[memory.number / 8]
        _psk = self._memobj.pskp_flags[memory.number / 8]

        bit = (1 << (memory.number % 8))

        if memory.empty:
            _usd |= bit
            self._set_bank(memory.number, None)
            return

        _usd &= ~bit

        _set_freq(_mem, memory.freq, memory.offset)
        _mem.name = memory.name.ljust(16)[:16]
        _mem.rtone = chirp_common.TONES.index(memory.rtone)
        _mem.ctone = chirp_common.TONES.index(memory.ctone)
        _mem.tmode = TMODES.index(memory.tmode)
        _mem.duplex = DUPLEX.index(memory.duplex)
        _mem.dtcs = chirp_common.DTCS_CODES.index(memory.dtcs)
        _mem.dtcs_polarity = DTCS_POLARITY.index(memory.dtcs_polarity)
        _mem.tune_step = TUNING_STEPS.index(memory.tuning_step)
        _mem.mode = next(i for i, mode in list(self.MODES.items())
                         if mode == memory.mode)

        if isinstance(memory, chirp_common.DVMemory):
            _mem.urcall = _encode_call(memory.dv_urcall.ljust(8))
            _mem.rpt1call = _encode_call(memory.dv_rpt1call.ljust(8))
            _mem.rpt2call = _encode_call(memory.dv_rpt2call.ljust(8))
            _mem.digcode = memory.dv_code
        elif memory.mode == "DV":
            LOG.debug('Converting Memory to DVMemory with default values')
            _mem.urcall = _encode_call('CQCQCQ'.ljust(8))
            _mem.rpt1call = _encode_call(' ' * 8)
            _mem.rpt2call = _encode_call(' ' * 8)
            _mem.digcode = 0

        if memory.skip == "S":
            _skp |= bit
            _psk &= ~bit
        elif memory.skip == "P":
            _skp |= bit
            _psk |= bit
        else:
            _skp &= ~bit
            _psk &= ~bit

        self._set_mem_extra(_mem, memory)

    def get_urcall_list(self):
        calls = []
        for i in range(0, 200):
            call = str(self._memobj.urcall[i].call)
            if call == "CALLSIGN":
                call = ""
            calls.append(call)
        return calls

    def get_mycall_list(self):
        calls = []
        for i in range(0, 6):
            calls.append(str(self._memobj.mycall[i].call))
        return calls

    def get_repeater_call_list(self):
        calls = []
        for rptcall in self._memobj.rptcall:
            call = _decode_call(rptcall.call)
            if call.rstrip() and not call == "CALLSIGN":
                calls.append(call)
        for repeater in self._memobj.repeaters:
            call = _decode_call(repeater.call)
            if call == "CALLSIGN":
                call = ""
            calls.append(call.rstrip())
        return calls


if __name__ == "__main__":
    print(repr(_decode_call(_encode_call("KD7REX B"))))
    print(repr(_decode_call(_encode_call("       B"))))
    print(repr(_decode_call(_encode_call("        "))))



#
# Copyright 2012 Filippi Marco <iz3gme.marco@gmail.com>
# Copyright 2018 Vinny Stipo <v@xpctech.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""FT818 management module"""

from chirp.drivers import ft817
from chirp import chirp_common, errors, directory
from chirp.settings import RadioSetting, RadioSettingValueBoolean
import logging

LOG = logging.getLogger(__name__)


@directory.register
class FT818Radio(ft817.FT817Radio):

    """Yaesu FT-818"""
    BAUD_RATE = 9600
    MODEL = "FT-818"
    _model = ""
    _memsize = 6573
    _block_lengths = [2, 40, 208, 208, 208, 208, 198, 53, 130, 118, 130]

    MEM_FORMAT = """
        struct mem_struct {
            u8  tag_on_off:1,
                tag_default:1,
                unknown1:3,
                mode:3;
            u8  duplex:2,
                is_duplex:1,
                is_cwdig_narrow:1,
                is_fm_narrow:1,
                freq_range:3;
            u8  skip:1,
                unknown2:1,
                ipo:1,
                att:1,
                unknown3:4;
            u8  ssb_step:2,
                am_step:3,
                fm_step:3;
            u8  unknown4:6,
                tmode:2;
            u8  unknown5:2,
                tx_mode:3,
                tx_freq_range:3;
            u8  unknown6:1,
                unknown_toneflag:1,
                tone:6;
            u8  unknown7:1,
                dcs:7;
            ul16 rit;
            u32 freq;
            u32 offset;
            u8  name[8];
        };

        #seekto 0x4;
        struct {
            u8  fst:1,
                lock:1,
                nb:1,
                pbt:1,
                unknownb:1,
                dsp:1,
                agc:2;
            u8  vox:1,
                vlt:1,
                bk:1,
                kyr:1,
                unknown5:1,
                cw_paddle:1,
                pwr_meter_mode:2;
            u8  vfob_band_select:4,
                vfoa_band_select:4;
            u8  unknowna;
            u8  backlight:2,
                color:2,
                contrast:4;
            u8  beep_freq:1,
                beep_volume:7;
            u8  arts_beep:2,
                main_step:1,
                cw_id:1,
                scope:1,
                pkt_rate:1,
                resume_scan:2;
            u8  op_filter:2,
                lock_mode:2,
                cw_pitch:4;
            u8  sql_rf_gain:1,
                ars_144:1,
                ars_430:1,
                cw_weight:5;
            u8  cw_delay;
            u8  unknown8:1,
                sidetone:7;
            u8  batt_chg:2,
                cw_speed:6;
            u8  disable_amfm_dial:1,
                vox_gain:7;
            u8  cat_rate:2,
                emergency:1,
                vox_delay:5;
            u8  dig_mode:3,
                mem_group:1,
                unknown9:1,
                apo_time:3;
            u8  dcs_inv:2,
                unknown10:1,
                tot_time:5;
            u8  mic_scan:1,
                ssb_mic:7;
            u8  mic_key:1,
                am_mic:7;
            u8  unknown11:1,
                fm_mic:7;
            u8  unknown12:1,
                dig_mic:7;
            u8  extended_menu:1,
                pkt_mic:7;
            u8  unknown14:1,
                pkt9600_mic:7;
            il16 dig_shift;
            il16 dig_disp;
            i8  r_lsb_car;
            i8  r_usb_car;
            i8  t_lsb_car;
            i8  t_usb_car;
            u8  unknown15:2,
                menu_item:6;
            u8  unknown16:4,
                menu_sel:4;
            u16 unknown17;
            u8  art:1,
                scn_mode:2,
                dw:1,
                pri:1,
                unknown18:1,
                tx_power:2;
            u8  spl:1,
                unknown:1,
                uhf_antenna:1,
                vhf_antenna:1,
                air_antenna:1,
                bc_antenna:1,
                sixm_antenna:1,
                hf_antenna:1;
        } settings;

        #seekto 0x2A;
        struct mem_struct vfoa[16];
        struct mem_struct vfob[16];
        struct mem_struct home[4];
        struct mem_struct qmb;
        struct mem_struct mtqmb;
        struct mem_struct mtune;

        #seekto 0x431;
        u8 visible[25];
        u8 pmsvisible;

        #seekto 0x44B;
        u8 filled[25];
        u8 pmsfilled;

        #seekto 0x465;
        struct mem_struct memory[200];
        struct mem_struct pms[2];

        #seekto 0x1903;
        u8 callsign[7];

        #seekto 0x19AD;
        struct mem_struct sixtymeterchannels[5];
    """

    SPECIAL_MEMORIES = {
        "VFOa-1.8M": -37,
        "VFOa-3.5M": -36,
        "VFOa-5M": -35,
        "VFOa-7M": -34,
        "VFOa-10M": -33,
        "VFOa-14M": -32,
        "VFOa-18M": -31,
        "VFOa-21M": -30,
        "VFOa-24M": -29,
        "VFOa-28M": -28,
        "VFOa-50M": -27,
        "VFOa-FM": -26,
        "VFOa-AIR": -25,
        "VFOa-144": -24,
        "VFOa-430": -23,
        "VFOa-HF": -22,
        "VFOb-1.8M": -21,
        "VFOb-3.5M": -20,
        "VFOb-5M": -19,
        "VFOb-7M": -18,
        "VFOb-10M": -17,
        "VFOb-14M": -16,
        "VFOb-18M": -15,
        "VFOb-21M": -14,
        "VFOb-24M": -13,
        "VFOb-28M": -12,
        "VFOb-50M": -11,
        "VFOb-FM": -10,
        "VFOb-AIR": -9,
        "VFOb-144": -8,
        "VFOb-430": -7,
        "VFOb-HF": -6,
        "HOME HF": -5,
        "HOME 50M": -4,
        "HOME 144": -3,
        "HOME 430": -2,
        "QMB": -1,
    }
    FIRST_VFOB_INDEX = -6
    LAST_VFOB_INDEX = -21
    FIRST_VFOA_INDEX = -22
    LAST_VFOA_INDEX = -37

    SPECIAL_PMS = {
        "PMS-L": -39,
        "PMS-U": -38,
    }
    LAST_PMS_INDEX = -39

    SPECIAL_MEMORIES.update(SPECIAL_PMS)

    SPECIAL_MEMORIES_REV = dict(zip(SPECIAL_MEMORIES.values(),
                                    SPECIAL_MEMORIES.keys()))


@directory.register
class FT818NDUSRadio(FT818Radio):

    """Yaesu FT-818ND (US version)"""
    MODEL = "FT-818ND (US)"

    _model = ""
    _US_model = True
    _memsize = 6703

    _block_lengths = [2, 40, 208, 208, 208, 208, 198, 53, 130, 118, 130, 130]

    SPECIAL_60M = {
        "M-601": -44,
        "M-602": -43,
        "M-603": -42,
        "M-604": -41,
        "M-605": -40,
    }
    LAST_SPECIAL60M_INDEX = -44

    SPECIAL_MEMORIES = dict(FT818Radio.SPECIAL_MEMORIES)
    SPECIAL_MEMORIES.update(SPECIAL_60M)

    SPECIAL_MEMORIES_REV = dict(zip(SPECIAL_MEMORIES.values(),
                                    SPECIAL_MEMORIES.keys()))

    def _get_special_60m(self, number):
        mem = chirp_common.Memory()
        mem.number = self.SPECIAL_60M[number]
        mem.extd_number = number

        _mem = self._memobj.sixtymeterchannels[-self.LAST_SPECIAL60M_INDEX +
                                               mem.number]

        mem = self._get_memory(mem, _mem)

        mem.immutable = ["number", "rtone", "ctone",
                         "extd_number", "name", "dtcs", "tmode", "cross_mode",
                         "dtcs_polarity", "power", "duplex", "offset",
                         "comment", "empty"]

        return mem

    def _set_special_60m(self, mem):
        if mem.empty:
            # can't delete 60M memories!
            raise Exception("Sorry, 60M memory can't be deleted")

        cur_mem = self._get_special_60m(self.SPECIAL_MEMORIES_REV[mem.number])

        for key in cur_mem.immutable:
            if cur_mem.__dict__[key] != mem.__dict__[key]:
                raise errors.RadioError("Editing field `%s' " % key +
                                        "is not supported on M-60x channels")

        if mem.mode not in ["USB", "LSB", "CW", "CWR", "NCW", "NCWR", "DIG"]:
            raise errors.RadioError("Mode {mode} is not valid "
                                    "in 60m channels".format(mode=mem.mode))
        _mem = self._memobj.sixtymeterchannels[-self.LAST_SPECIAL60M_INDEX +
                                               mem.number]
        self._set_memory(mem, _mem)

    def get_memory(self, number):
        if number in self.SPECIAL_60M.keys():
            return self._get_special_60m(number)
        elif number < 0 and \
                self.SPECIAL_MEMORIES_REV[number] in self.SPECIAL_60M.keys():
            # I can't stop delete operation from losing extd_number but
            # I know how to get it back
            return self._get_special_60m(self.SPECIAL_MEMORIES_REV[number])
        else:
            return FT818Radio.get_memory(self, number)

    def set_memory(self, memory):
        if memory.number in self.SPECIAL_60M.values():
            return self._set_special_60m(memory)
        else:
            return FT818Radio.set_memory(self, memory)

    def get_settings(self):
        top = FT818Radio.get_settings(self)
        basic = top[0]
        rs = RadioSetting("emergency", "Emergency",
                          RadioSettingValueBoolean(
                              self._memobj.settings.emergency))
        basic.append(rs)
        return top



# Copyright 2022 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import logging
import os
import serial
import sys
import tempfile

import requests
import wx
import wx.richtext
import wx.lib.scrolledpanel

from chirp import bitwise
from chirp import util
from chirp.wxui import common
from chirp.wxui import report

LOG = logging.getLogger(__name__)
BrowserChanged, EVT_BROWSER_CHANGED = wx.lib.newevent.NewCommandEvent()
FROZEN = getattr(sys, 'frozen', False)


def simple_diff(a, b, diffsonly=False):
    lines_a = a.split(os.linesep)
    lines_b = b.split(os.linesep)
    blankprinted = True

    diff = ""
    for i in range(0, len(lines_a)):
        if lines_a[i] != lines_b[i]:
            diff += "-%s%s" % (lines_a[i], os.linesep)
            diff += "+%s%s" % (lines_b[i], os.linesep)
            blankprinted = False
        elif diffsonly is True:
            if blankprinted:
                continue
            diff += os.linesep
            blankprinted = True
        else:
            diff += " %s%s" % (lines_a[i], os.linesep)
    return diff


class MemoryDialog(wx.Dialog):
    def __init__(self, mem, *a, **k):

        super(MemoryDialog, self).__init__(*a, **k)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        self.text = wx.richtext.RichTextCtrl(
            self, style=wx.VSCROLL | wx.HSCROLL | wx.NO_BORDER)
        sizer.Add(self.text, 1, wx.EXPAND)

        sizer.Add(wx.StaticLine(self), 0)

        bs = self.CreateButtonSizer(wx.OK | wx.OK_DEFAULT)
        sizer.Add(bs)

        self.Bind(wx.EVT_BUTTON, lambda e: self.EndModal(wx.CLOSE))

        font = wx.Font(12, wx.FONTFAMILY_MODERN, wx.FONTSTYLE_NORMAL,
                       wx.FONTWEIGHT_NORMAL)

        self.text.BeginFont(font)

        if isinstance(mem, tuple):
            mem_a, mem_b = mem
            self._diff_memories(mem_a, mem_b)
            self.SetTitle(_('Diff Raw Memories'))
        else:
            self._raw_memory(mem)
            self.SetTitle(_('Show Raw Memory'))

        self.SetSize(640, 480)
        self.Centre()

    def _raw_memory(self, mem):
        self.text.WriteText(mem)

    def _diff_memories(self, mem_a, mem_b):
        diff = simple_diff(mem_a, mem_b)

        for line in diff.split(os.linesep):
            color = None
            if line.startswith('+'):
                self.text.BeginTextColour((255, 0, 0))
                color = True
            elif line.startswith('-'):
                self.text.BeginTextColour((0, 0, 255))
                color = True
            self.text.WriteText(line)
            self.text.Newline()
            if color:
                self.text.EndTextColour()


class ChirpEditor(wx.Panel):
    def __init__(self, parent, obj):
        super(ChirpEditor, self).__init__(parent, )
        self._obj = obj
        self._fixed_font = wx.Font(pointSize=10,
                                   family=wx.FONTFAMILY_TELETYPE,
                                   style=wx.FONTSTYLE_NORMAL,
                                   weight=wx.FONTWEIGHT_NORMAL)
        self._changed_color = wx.Colour(0, 255, 0)
        self._error_color = wx.Colour(255, 0, 0)

    def refresh(self):
        """Called to refresh the widget from memory"""
        pass

    def set_up(self):
        pass

    def _mark_changed(self, thing):
        thing.SetBackgroundColour(self._changed_color)
        tt = thing.GetToolTip()
        if not tt:
            tt = wx.ToolTip('')
            thing.SetToolTip(tt)
        tt.SetTip(_('Press enter to set this in memory'))

    def _mark_unchanged(self, thing, mem_changed=True):
        thing.SetBackgroundColour(wx.NullColour)
        thing.UnsetToolTip()
        if mem_changed:
            wx.PostEvent(self, BrowserChanged(self.GetId()))

    def _mark_error(self, thing, reason):
        tt = thing.GetToolTip()
        if not tt:
            tt = wx.ToolTip('')
            thing.SetToolTip(tt)
        tt.SetTip(reason)
        thing.SetBackgroundColour(self._error_color)

    def __repr__(self):
        addr = '0x%02x' % int(self._obj._offset)

        def typestr(c):
            return c.__class__.__name__.lower().replace('dataelement', '')

        if isinstance(self._obj, bitwise.arrayDataElement):
            innertype = list(self._obj.items())[0][1]
            return '%s[%i] (%i %s) @ %s' % (typestr(innertype),
                                            len(self._obj),
                                            innertype.size() / 8,
                                            _('bytes each'),
                                            addr)
        elif self._obj.size() % 8 == 0:
            return '%s (%i %s) @ %s' % (typestr(self._obj),
                                        self._obj.size() / 8,
                                        _('bytes'),
                                        addr)
        else:
            return '%s %s @ %s' % (self._obj.size(),
                                   _('bits'),
                                   addr)


class ChirpStringEditor(ChirpEditor):
    def set_up(self):
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._entry = wx.TextCtrl(self, value=str(self._obj),
                                  style=wx.TE_PROCESS_ENTER)
        self._entry.SetMaxLength(len(self._obj))
        self.SetSizer(sizer)
        sizer.Add(self._entry, 1, wx.EXPAND)

        self._entry.Bind(wx.EVT_TEXT, self._edited)
        self._entry.Bind(wx.EVT_TEXT_ENTER, self._changed)
        self._entry.SetEditable(not FROZEN)

    def refresh(self):
        self._entry.SetValue(str(self._obj))
        self._mark_unchanged(self._entry, mem_changed=False)

    def _edited(self, event):
        entry = event.GetEventObject()
        value = entry.GetValue()
        if len(value) == len(self._obj):
            self._mark_changed(entry)
        else:
            self._mark_error(entry, _('Length must be %i') % len(self._obj))

    @common.error_proof()
    def _changed(self, event):
        entry = event.GetEventObject()
        value = entry.GetValue()
        self._obj.set_value(value)
        self._mark_unchanged(entry)
        LOG.debug('Set value: %r' % value)


class ChirpIntegerEditor(ChirpEditor):
    def set_up(self):
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(sizer)

        hexdigits = (self._obj.size() / 4) + (self._obj.size() % 4 and 1 or 0)
        bindigits = self._obj.size()

        self._editors = {_('Hex'): (16, '{:0%iX}' % hexdigits),
                         _('Dec'): (10, '{:d}'),
                         _('Bin'): (2, '{:0%ib}' % bindigits)}
        self._entries = {}
        for name, (base, fmt) in self._editors.items():
            label = wx.StaticText(self, label=name)
            entry = wx.TextCtrl(self, value=fmt.format(int(self._obj)),
                                style=wx.TE_PROCESS_ENTER)
            entry.SetFont(self._fixed_font)
            entry.SetEditable(not FROZEN)
            sizer.Add(label, 0, wx.ALIGN_CENTER)
            sizer.Add(entry, 1, flag=wx.EXPAND)
            self._entries[name] = entry

            entry.Bind(wx.EVT_TEXT, functools.partial(self._edited,
                                                      base=base))
            entry.Bind(wx.EVT_TEXT_ENTER, functools.partial(self._changed,
                                                            base=base))

    def refresh(self):
        for name, (base, fmt) in self._editors.items():
            self._entries[name].SetValue(fmt.format(int(self._obj)))
            self._mark_unchanged(self._entries[name],
                                 mem_changed=False)

    def _edited(self, event, base=10):
        entry = event.GetEventObject()
        others = {n: e for n, e in self._entries.items()
                  if e != entry}

        try:
            val = int(entry.GetValue(), base)
            assert val >= 0, _('Value must be zero or greater')
            assert val < pow(2, self._obj.size()), \
                _('Value does not fit in %i bits') % self._obj.size()
        except (ValueError, AssertionError) as e:
            self._mark_error(entry, str(e))
            return
        else:
            self._mark_changed(entry)

        for name, entry in others.items():
            base, fmt = self._editors[name]
            entry.ChangeValue(fmt.format(val))

    @common.error_proof()
    def _changed(self, event, base=10):
        entry = event.GetEventObject()
        val = int(entry.GetValue(), base)
        self._obj.set_value(val)
        self._mark_unchanged(entry)
        LOG.debug('Set value: %r' % val)


class ChirpBCDEditor(ChirpEditor):
    def set_up(self):
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(sizer)
        self._entry = wx.TextCtrl(self, value=str(int(self._obj)),
                                  style=wx.TE_PROCESS_ENTER)
        self._entry.SetFont(self._fixed_font)
        self._entry.SetEditable(not FROZEN)
        sizer.Add(self._entry, 1, wx.EXPAND)
        self._entry.Bind(wx.EVT_TEXT, self._edited)
        self._entry.Bind(wx.EVT_TEXT_ENTER, self._changed)

    def refresh(self):
        self._entry.SetValue(str(int(self._obj)))
        self._mark_unchanged(self._entry,
                             mem_changed=False)

    def _edited(self, event):
        entry = event.GetEventObject()
        try:
            val = int(entry.GetValue())
            digits = self._obj.size() // 4
            assert val >= 0, _('Value must be zero or greater')
            assert len(entry.GetValue()) == digits, \
                _('Value must be exactly %i decimal digits') % digits
        except (ValueError, AssertionError) as e:
            self._mark_error(entry, str(e))
        else:
            self._mark_changed(entry)

    @common.error_proof()
    def _changed(self, event):
        entry = event.GetEventObject()
        val = int(entry.GetValue())
        self._obj.set_value(val)
        self._mark_unchanged(entry)
        LOG.debug('Set Value: %r' % val)


class ChirpBrowserPanel(wx.lib.scrolledpanel.ScrolledPanel):
    def __init__(self, parent, memobj):
        super(ChirpBrowserPanel, self).__init__(parent)
        self._sizer = wx.FlexGridSizer(2)
        self._sizer.AddGrowableCol(1)
        self.SetSizer(self._sizer)
        self.SetupScrolling()
        self._parent = parent
        self._memobj = memobj
        self._editors = {}
        self._initialized = False

    def add_editor(self, name, editor):
        self._editors[name] = editor
        editor.Bind(EVT_BROWSER_CHANGED, self._panel_changed)

    def _panel_changed(self, event):
        wx.PostEvent(self, BrowserChanged(self.GetId()))

    def _initialize(self):
        for name, obj in self._memobj.items():
            editor = None
            if isinstance(obj, bitwise.arrayDataElement):
                if isinstance(obj[0], bitwise.charDataElement):
                    editor = ChirpStringEditor(self, obj)
                elif isinstance(obj[0], bitwise.bcdDataElement):
                    editor = ChirpBCDEditor(self, obj)
                else:
                    self._parent.add_sub_panel(name, obj, self)
            elif isinstance(obj, bitwise.intDataElement):
                editor = ChirpIntegerEditor(self, obj)
            elif isinstance(obj, bitwise.structDataElement):
                self._parent.add_sub_panel(name, obj, self)
            if editor:
                self.add_editor(name, editor)
        self._initialized = True

    def add_sub_panel(self, name, obj, parent):
        # This just forwards until we get to the browser
        self._parent.add_sub_panel(name, obj, parent)

    def selected(self):
        if not self._initialized:
            self._initialize()

            label = wx.StaticText(self)
            pos = wx.StaticText(self, label='%i bits (%i bytes) at 0x%06x' % (
                    self._memobj.size(),
                    self._memobj.size() // 8,
                    self._memobj.get_offset()))
            self._sizer.Add(label, 0, wx.ALIGN_CENTER)
            self._sizer.Add(pos, 1, flag=wx.EXPAND)

            for name, editor in self._editors.items():
                editor.set_up()
                label = wx.StaticText(self, label='%s: ' % name)
                tt = wx.ToolTip(repr(editor))
                label.SetToolTip(tt)

                self._sizer.Add(label, 0, wx.ALIGN_CENTER)
                self._sizer.Add(editor, 1, flag=wx.EXPAND)
        else:
            for editor in self._editors.values():
                editor.refresh()

        self._sizer.Layout()
        self.FitInside()


class ChirpRadioBrowser(common.ChirpEditor, common.ChirpSyncEditor):
    def __init__(self, radio, *a, **k):
        super(ChirpRadioBrowser, self).__init__(*a, **k)
        self._loaded = False
        self._radio = radio
        self._features = radio.get_features()

        self._treebook = ChirpBrowserTreeBook(self)

        try:
            view = self._treebook.GetTreeCtrl()
        except AttributeError:
            # https://github.com/wxWidgets/Phoenix/issues/918
            view = self._treebook.GetChildren()[0]

        view.SetMinSize((250, 0))

        self._treebook.Bind(wx.EVT_TREEBOOK_PAGE_CHANGED,
                            self.page_selected)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self._treebook, 1, wx.EXPAND)
        self.SetSizer(sizer)

    def selected(self):
        if self._loaded:
            self._treebook.CurrentPage.selected()
            return

        self.start_wait_dialog(_('Building Radio Browser'))
        self._loaded = True
        try:
            self._treebook.add_sub_panel('%s %s' % (self._radio.VENDOR,
                                                    self._radio.MODEL),
                                         self._radio._memobj, self._treebook)
        except Exception as e:
            LOG.exception('Failed to load browser: %s' % e)
            common.error_proof.show_error(_('Failed to load radio browser'))
        finally:
            self.stop_wait_dialog()
        if self._treebook.GetPageCount():
            self._treebook.ExpandNode(0)

    def page_selected(self, event):
        page = self._treebook.GetPage(event.GetSelection())
        page.selected()


class ChirpBrowserTreeBook(wx.Treebook):
    def add_sub_panel(self, name, memobj, parent):
        LOG.debug('Adding sub panel for %s' % name)
        page = ChirpBrowserPanel(self, memobj)
        page.Bind(EVT_BROWSER_CHANGED, self._page_changed)
        if parent != self:
            pos = self.FindPage(parent)
            self.InsertSubPage(pos, page, name)
            self.ExpandNode(pos)
        else:
            self.AddPage(page, name)

    def _page_changed(self, event):
        wx.PostEvent(self, common.EditorChanged(self.GetId()))


class FakeSerial(serial.SerialBase):
    def write(self, buf):
        LOG.debug('Fake serial write:\n%s' % util.hexprint(buf))

    def read(self, count):
        LOG.debug('Fake serial read %i' % count)
        return b''


class FakeEchoSerial(FakeSerial):
    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self._buf = []

    def write(self, buf):
        super().write(buf)
        self._buf.extend(b for b in buf)

    def read(self, count):
        super().read(count)
        try:
            return bytes([self._buf.pop(0)])
        except IndexError:
            LOG.warning('Empty echo buffer')
            return b''


class IssueModuleLoader:
    def __init__(self, parent):
        self._parent = parent
        report.ensure_session()
        self.session = report.SESSION

    def get_attachments_from_issue(self, issue):
        r = self.session.get(
            'https://chirp.danplanet.com/issues/%i.json' % issue,
            params={'include': 'attachments'})
        LOG.debug('Fetched attachments for issue %i (status %s)' % (
            issue, r.status_code))
        r.raise_for_status()
        data = r.json()['issue']['attachments']
        return [a for a in data if
                a['filename'].endswith('.py') and
                a.get('content_type', '').startswith('text/') and
                a['filesize'] < (256 * 1024)]

    def get_user_is_developer(self, uid):
        r = self.session.get('https://chirp.danplanet.com/users/%i.json' % uid,
                             params={'include': 'memberships'})
        LOG.debug('Fetched info for user %i (status %s)',
                  uid, r.status_code)
        r.raise_for_status()
        data = r.json()
        try:
            membership = data['user']['memberships'][0]
        except IndexError:
            LOG.debug('User %s(%i) has no roles', data['user']['login'], uid)
            return False
        roles = [r['name'] for r in membership['roles']]
        return 'Developer' in roles or 'Manager' in roles

    def get_attachment_from_user(self, issue, attachments):
        attachment_strings = {
            '%s from %s (%s)' % (a['filename'],
                                 a['author']['name'],
                                 a['created_on']): a
            for a in sorted(attachments, key=lambda a: a['created_on'])
            }
        choices = list(attachment_strings.keys())
        choice = wx.GetSingleChoice(
            _('Choose the module to load from issue %i:' % issue),
            _('Available modules'),
            choices,
            len(choices) - 1,
            parent=self._parent)
        if choice:
            isdev = self.get_user_is_developer(
                attachment_strings[choice]['author']['id'])
            if not isdev:
                r = wx.MessageBox(
                    _('The author of this module is not a recognized '
                      'CHIRP developer. It is recommended that you not '
                      'load this module as it could pose a security risk. '
                      'Proceed anyway?'), _('Security Risk'),
                    wx.YES_NO | wx.NO_DEFAULT)
                if r != wx.YES:
                    return
            return attachment_strings[choice]

    def run(self):
        msg = _('This will load a module from a website issue')
        issue = wx.GetNumberFromUser(msg,
                                     _('Issue number:'),
                                     _('Load module from issue'),
                                     0, 0, 999999, parent=self._parent)
        if issue < 0:
            return

        try:
            attachments = self.get_attachments_from_issue(issue)
        except Exception as e:
            LOG.exception('Failed to load attachments from %i: %s' % (
                issue, e))
            raise Exception('Unable to load modules from that issue')
        LOG.debug('Found %i valid module attachments from issue %i' % (
            len(attachments), issue))
        if not attachments:
            wx.MessageBox(_('No modules found in issue %i') % issue,
                          _('No modules found'),
                          wx.ICON_WARNING)
            return
        attachment = self.get_attachment_from_user(issue, attachments)
        if not attachment:
            return
        LOG.debug('User chose attachment %s' % attachment)

        LOG.debug('Fetching attachment URL %s' % attachment['content_url'])
        r = requests.get(attachment['content_url'])
        modfile = tempfile.mktemp('.py', 'loaded-%i-' % attachment['id'])
        trailer = ('\n\n# Loaded from issue %i attachment %i: %s\n' % (
            issue, attachment['id'], attachment['content_url']))
        with open(modfile, 'wb') as f:
            f.write(r.content)
            f.write(trailer.encode())

        LOG.debug('Wrote attachment to %s' % modfile)
        return modfile



# Copyright 2016 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time

from chirp import chirp_common, directory, errors
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettingValueString, \
    RadioSettings


LOG = logging.getLogger(__name__)
BAUDS = [1200, 4800, 9600, 19200, 38400, 57600]
POWER_LEVELS = [chirp_common.PowerLevel('Low', dBm=10),
                chirp_common.PowerLevel('High', dBm=24)]
TONE_MODES = ['', 'Tone', 'TSQL', '']


def detect_baudrate(radio):
    bauds = list(BAUDS)
    bauds.remove(radio.pipe.baudrate)
    bauds.insert(0, radio.pipe.baudrate)
    for baud in bauds:
        radio.pipe.baudrate = baud
        radio.pipe.timeout = 0.5
        radio.pipe.write(b'\rFW?\r')
        resp = radio.pipe.read(2)
        if resp.strip().startswith(b'FW'):
            resp += radio.pipe.read(16)
            LOG.info('HobbyPCB %r at baud rate %i' % (resp.strip(), baud))
            return baud


@directory.register
class HobbyPCBRSUV3Radio(chirp_common.LiveRadio):
    """HobbyPCB RS-UV3"""
    VENDOR = "HobbyPCB"
    MODEL = "RS-UV3"
    BAUD_RATE = 19200
    NEEDS_COMPAT_SERIAL = False

    def __init__(self, *args, **kwargs):
        super(HobbyPCBRSUV3Radio, self).__init__(*args, **kwargs)
        if self.pipe:
            baud = detect_baudrate(self)
            if not baud:
                raise errors.RadioError('Radio did not respond')

    def _cmd(self, command, rsize=None):
        command = command.encode()
        LOG.debug('> %s' % command)
        self.pipe.write(b'%s\r' % command)
        resp = b''

        if rsize is None:
            complete = lambda: time.sleep(0.1) is None
        elif rsize == 0:
            rsize = 1
            complete = lambda: resp.endswith(b'\r')
        else:
            complete = lambda: len(resp) >= rsize

        while not complete():
            chunk = self.pipe.read(rsize)
            if not chunk:
                break
            resp += chunk
        LOG.debug('< %r [%i]' % (resp, len(resp)))
        return resp.decode().strip()

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_bank = False
        rf.has_name = False
        rf.has_cross = False
        rf.has_dtcs = False
        rf.has_rx_dtcs = False
        rf.has_dtcs_polarity = False
        rf.has_tuning_step = False
        rf.has_mode = False
        rf.has_settings = True
        rf.memory_bounds = (1, 9)  # This radio supports memories 0-9
        rf.valid_bands = [(144000000, 148000000),
                          (220000000, 222000000),
                          (440000000, 450000000),
                          ]
        rf.valid_tmodes = TONE_MODES
        rf.valid_power_levels = POWER_LEVELS
        return rf

    def get_memory(self, number):
        _mem = self._cmd('CP%i' % number, 33).split('\r')
        LOG.debug('Memory elements: %r' % _mem)
        mem = chirp_common.Memory()
        mem.number = number
        mem.freq = int(_mem[0]) * 1000
        txfreq = int(_mem[1]) * 1000
        mem.offset = abs(txfreq - mem.freq)
        if mem.freq < txfreq:
            mem.duplex = '+'
        elif mem.freq > txfreq:
            mem.duplex = '-'
        else:
            mem.duplex = ''
        mem.ctone = int(_mem[2]) / 100.0
        mem.rtone = mem.ctone
        mem.tmode = TONE_MODES[int(_mem[3])]
        mem.power = POWER_LEVELS[int(_mem[5])]
        return mem

    def set_memory(self, mem):
        if mem.tmode in ['', 'Tone']:
            tone = mem.rtone * 100
        else:
            tone = mem.ctone * 100
        if mem.duplex == '+':
            self._cmd('FT%06i' % ((mem.freq + mem.offset) / 1000))
            self._cmd('FR%06i' % (mem.freq / 1000))
        elif mem.duplex == '-':
            self._cmd('FT%06i' % ((mem.freq - mem.offset) / 1000))
            self._cmd('FR%06i' % (mem.freq / 1000))
        else:
            self._cmd('FS%06i' % (mem.freq / 1000))
        self._cmd('TM%i' % TONE_MODES.index(mem.tmode))
        self._cmd('TF%05i' % tone)
        self._cmd('PW%i' % POWER_LEVELS.index(mem.power))
        self._cmd('ST%i' % mem.number)

    def get_settings(self):
        def _get(cmd):
            return self._cmd('%s?' % cmd, 0).split(':')[1].strip()

        cw = RadioSettingGroup('beacon', 'Beacon Settings')
        cl = RadioSetting('CL%15s', 'CW Callsign',
                          RadioSettingValueString(0, 15,
                                                  _get('CL')))
        cw.append(cl)

        cf = RadioSetting('CF%4i', 'CW Audio Frequency',
                          RadioSettingValueInteger(400, 1300,
                                                   int(_get('CF'))))
        cw.append(cf)

        cs = RadioSetting('CS%02i', 'CW Speed',
                          RadioSettingValueInteger(5, 25,
                                                   int(_get('CS'))))
        cw.append(cs)

        bc = RadioSetting('BC%03i', 'CW Beacon Timer',
                          RadioSettingValueInteger(0, 600,
                                                   int(_get('BC'))))
        cw.append(bc)

        bm = RadioSetting('BM%15s', 'Beacon Message',
                          RadioSettingValueString(0, 15,
                                                  _get('BM')))
        cw.append(bm)

        bt = RadioSetting('BT%03i', 'Beacon Timer',
                          RadioSettingValueInteger(0, 600,
                                                   int(_get('BT'))))
        cw.append(bt)

        it = RadioSetting('IT%03i', 'CW ID Timer',
                          RadioSettingValueInteger(0, 500,
                                                   int(_get('IT'))))
        cw.append(it)

        tg = RadioSetting('TG%7s', 'CW Timeout Message',
                          RadioSettingValueString(0, 7,
                                                  _get('TG')))
        cw.append(tg)

        io = RadioSettingGroup('io', 'IO')

        af = RadioSetting('AF%i', 'Arduino LPF',
                          RadioSettingValueBoolean(_get('AF') == 'ON'))
        io.append(af)

        input_pin = ['OFF', 'SQ OPEN', 'PTT']
        ai = RadioSetting('AI%i', 'Arduino Input Pin',
                          RadioSettingValueList(
                              input_pin,
                              input_pin[int(_get('AI'))]))
        io.append(ai)

        output_pin = ['LOW', 'SQ OPEN', 'DTMF DETECT', 'TX ON', 'CTCSS DET',
                      'HIGH']
        ao = RadioSetting('AO%i', 'Arduino Output Pin',
                          RadioSettingValueList(
                              output_pin,
                              output_pin[int(_get('AO'))]))
        io.append(ao)

        bauds = [str(x) for x in BAUDS]
        b1 = RadioSetting('B1%i', 'Arduino Baudrate',
                          RadioSettingValueList(
                              bauds,
                              bauds[int(_get('B1'))]))
        io.append(b1)

        b2 = RadioSetting('B2%i', 'Main Baudrate',
                          RadioSettingValueList(
                              bauds,
                              bauds[int(_get('B2'))]))
        io.append(b2)

        dtmf = RadioSettingGroup('dtmf', 'DTMF Settings')

        dd = RadioSetting('DD%04i', 'DTMF Tone Duration',
                          RadioSettingValueInteger(50, 2000,
                                                   int(_get('DD'))))
        dtmf.append(dd)

        dr = RadioSetting('DR%i', 'DTMF Tone Detector',
                          RadioSettingValueBoolean(_get('DR') == 'ON'))
        dtmf.append(dr)

        gt = RadioSetting('GT%02i', 'DTMF/CW Tone Gain',
                          RadioSettingValueInteger(0, 15,
                                                   int(_get('GT'))))
        dtmf.append(gt)

        sd = RadioSetting('SD%i', 'DTMF/CW Side Tone',
                          RadioSettingValueBoolean(_get('SD') == 'ON'))
        dtmf.append(sd)

        general = RadioSettingGroup('general', 'General')

        dp = RadioSetting('DP%i', 'Pre-Emphasis',
                          RadioSettingValueBoolean(_get('DP') == 'ON'))
        general.append(dp)

        fw = RadioSetting('_fw', 'Firmware Version',
                          RadioSettingValueString(0, 20,
                                                  _get('FW')))
        general.append(fw)

        gm = RadioSetting('GM%02i', 'Mic Gain',
                          RadioSettingValueInteger(0, 15,
                                                   int(_get('GM'))))
        general.append(gm)

        hp = RadioSetting('HP%i', 'Audio High-Pass Filter',
                          RadioSettingValueBoolean(_get('HP') == 'ON'))
        general.append(hp)

        ht = RadioSetting('HT%04i', 'Hang Time',
                          RadioSettingValueInteger(0, 5000,
                                                   int(_get('HT'))))
        general.append(ht)

        ledmode = ['OFF', 'ON', 'SQ OPEN', 'BATT CHG STAT']
        ld = RadioSetting('LD%i', 'LED Mode',
                          RadioSettingValueList(
                              ledmode,
                              ledmode[int(_get('LD'))]))
        general.append(ld)

        sq = RadioSetting('SQ%i', 'Squelch Level',
                          RadioSettingValueInteger(0, 9,
                                                   int(_get('SQ'))))
        general.append(sq)

        to = RadioSetting('TO%03i', 'Timeout Timer',
                          RadioSettingValueInteger(0, 600,
                                                   int(_get('TO'))))
        general.append(to)

        vu = RadioSetting('VU%02i', 'Receiver Audio Volume',
                          RadioSettingValueInteger(0, 39,
                                                   int(_get('VU'))))
        general.append(vu)

        rc = RadioSetting('RC%i', 'Current Channel',
                          RadioSettingValueInteger(0, 9, 0))
        rc.set_doc('Choosing one of these values causes the radio '
                   'to change to the selected channel. The radio '
                   'cannot tell CHIRP what channel is selected.')
        general.append(rc)

        return RadioSettings(general, cw, io, dtmf)

    def set_settings(self, settings):
        def _set(thing):
            # Try to only set something if it's new
            query = '%s?' % thing[:2]
            cur = self._cmd(query, 0)
            if cur.strip():
                cur = cur.split()[1].strip()
            new = thing[2:].strip()
            if cur in ['ON', 'OFF']:
                cur = int(cur == 'ON')
                new = int(new)
            elif cur.isdigit():
                cur = int(cur)
                new = int(new)
            if new != cur:
                LOG.info('Setting %s (%r != %r)' % (thing, cur, new))
                self._cmd(thing)
                time.sleep(1)

        for group in settings:
            for setting in group:
                if setting.get_name().startswith('_'):
                    LOG.debug('Skipping %s' % setting)
                    continue
                cmd = setting.get_name()
                value = setting.value.get_value()
                if hasattr(setting.value, '_options'):
                    value = setting.value._options.index(value)
                fullcmd = (cmd % value).strip()
                _set(fullcmd)
